from spaceone.dashboard.model.public_folder.database import PublicFolder
from spaceone.dashboard.model.public_dashboard.database import PublicDashboard
from spaceone.dashboard.model.public_widget.database import PublicWidget
from spaceone.dashboard.model.public_data_table.database import PublicDataTable
from spaceone.dashboard.model.private_folder.database import PrivateFolder
from spaceone.dashboard.model.private_dashboard.database import PrivateDashboard
from spaceone.dashboard.model.private_widget.database import PrivateWidget
from spaceone.dashboard.model.private_data_table.database import PrivateDataTable



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.manager.private_dashboard_manager import PrivateDashboardManager
from spaceone.dashboard.manager.private_folder_manager import PrivateFolderManager
from spaceone.dashboard.manager.identity_manager import IdentityManager
from spaceone.dashboard.model.private_dashboard.request import *
from spaceone.dashboard.model.private_dashboard.response import *
from spaceone.dashboard.model.private_dashboard.database import PrivateDashboard
from spaceone.dashboard.service.private_widget_service import PrivateWidgetService

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PrivateDashboardService(BaseService):
    resource = "PrivateDashboard"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pri_dashboard_mgr = PrivateDashboardManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="dashboard:PrivateDashboard.write",
        role_types=["USER"],
    )
    @convert_model
    def create(
        self, params: PrivateDashboardCreateRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Create private dashboard

        Args:
            params (dict): {
                'name': 'str',                  # required
                'description': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'vars_schema': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'dict',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'workspace_id': 'str',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        pri_dashboard_vo = self.create_dashboard(params.dict())
        return PrivateDashboardResponse(**pri_dashboard_vo)

    @check_required(["name", "domain_id", "user_id"])
    def create_dashboard(self, params_dict: dict) -> dict:
        domain_id = params_dict["domain_id"]
        user_id = params_dict["user_id"]
        workspace_id = params_dict.get("workspace_id")

        layouts = params_dict.get("layouts")
        if layouts:
            del params_dict["layouts"]

        if folder_id := params_dict.get("folder_id"):
            pri_folder_mgr = PrivateFolderManager()
            pri_folder_mgr.get_private_folder(folder_id, domain_id, user_id)

        if workspace_id:
            self.identity_mgr.check_workspace(workspace_id, domain_id)

        pri_dashboard_vo = self.pri_dashboard_mgr.create_private_dashboard(params_dict)

        if layouts:
            pri_widget_svc = PrivateWidgetService()
            created_layouts = []
            for layout in layouts:
                widgets = layout.get("widgets", [])
                created_layout = {"name": layout.get("name", ""), "widgets": []}
                for widget in widgets:
                    widget["state"] = "ACTIVE"
                    widget["dashboard_id"] = pri_dashboard_vo.dashboard_id
                    widget["domain_id"] = domain_id
                    widget["user_id"] = user_id
                    widget["is_bulk"] = True

                    widget_info = pri_widget_svc.create_widget(widget)
                    created_widget_id = widget_info["widget_id"]

                    _LOGGER.debug(
                        f"[create_dashboard] create widget: {created_widget_id}"
                    )
                    created_layout["widgets"].append(created_widget_id)

                created_layouts.append(created_layout)

            self.pri_dashboard_mgr.update_private_dashboard_by_vo(
                {"layouts": created_layouts}, pri_dashboard_vo
            )

        return pri_dashboard_vo.to_dict()

    @transaction(
        permission="dashboard:PrivateDashboard.write",
        role_types=["USER"],
    )
    @convert_model
    def update(
        self, params: PrivateDashboardUpdateRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Update private dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'name': 'str',
                'description': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'vars_schema': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'list',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        pri_dashboard_vo: PrivateDashboard = (
            self.pri_dashboard_mgr.get_private_dashboard(
                params.dashboard_id, params.domain_id, params.user_id
            )
        )

        if params.folder_id:
            pri_folder_mgr = PrivateFolderManager()
            pri_folder_mgr.get_private_folder(
                params.folder_id, params.domain_id, params.user_id
            )

        pri_dashboard_vo = self.pri_dashboard_mgr.update_private_dashboard_by_vo(
            params.dict(exclude_unset=True), pri_dashboard_vo
        )

        return PrivateDashboardResponse(**pri_dashboard_vo.to_dict())

    @transaction(permission="dashboard:PrivateDashboard.write", role_types=["USER"])
    @convert_model
    def change_folder(
        self, params: PrivateDashboardChangeFolderRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Change private dashboard's folder

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'folder_id': 'str',      # required
                'user_id': 'str',        # injected from auth (required)
                'domain_id': 'str'       # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        folder_id = params.folder_id
        pri_dashboard_vo = self.pri_dashboard_mgr.get_private_dashboard(
            params.dashboard_id, params.domain_id, params.user_id
        )

        if folder_id:
            pri_folder_mgr = PrivateFolderManager()
            pri_folder_mgr.get_private_folder(
                params.folder_id, params.domain_id, params.user_id
            )

        pri_dashboard_vo = self.pri_dashboard_mgr.update_private_dashboard_by_vo(
            params.dict(), pri_dashboard_vo
        )

        return PrivateDashboardResponse(**pri_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDashboard.write",
        role_types=["USER"],
    )
    @convert_model
    def delete(self, params: PrivateDashboardDeleteRequest) -> None:
        """Delete private dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_dashboard_vo = self.pri_dashboard_mgr.get_private_dashboard(
            params.dashboard_id, params.domain_id, params.user_id
        )

        self.pri_dashboard_mgr.delete_private_dashboard_by_vo(pri_dashboard_vo)

    @transaction(
        permission="dashboard:PrivateDashboard.read",
        role_types=["USER"],
    )
    @convert_model
    def get(
        self, params: PrivateDashboardGetRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Get private dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        pri_dashboard_vo = self.pri_dashboard_mgr.get_private_dashboard(
            params.dashboard_id, params.domain_id, params.user_id
        )

        return PrivateDashboardResponse(**pri_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDashboard.read",
        role_types=["USER"],
    )
    @append_query_filter(
        ["dashboard_id", "name", "domain_id", "workspace_id", "user_id", "folder_id"]
    )
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def list(
        self, params: PrivateDashboardSearchQueryRequest
    ) -> Union[PrivateDashboardsResponse, dict]:
        """List private dashboards

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',
                'name': 'str',
                'folder_id': 'str',
                'user_id': 'str',                               # injected from auth (required)
                'workspace_id': 'str',
                'domain_id': 'str',                             # injected from auth (required)
            }

        Returns:
            PrivateDashboardsResponse:
        """

        query = params.query or {}
        pri_dashboard_vos, total_count = self.pri_dashboard_mgr.list_private_dashboards(
            query
        )
        pri_dashboards_info = [
            pri_dashboard_vo.to_dict() for pri_dashboard_vo in pri_dashboard_vos
        ]
        return PrivateDashboardsResponse(
            results=pri_dashboards_info, total_count=total_count
        )

    @transaction(
        permission="dashboard:PrivateDashboard.read",
        role_types=["USER"],
    )
    @append_query_filter(["domain_id", "user_id"])
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def stat(self, params: PrivateDashboardStatQueryRequest) -> dict:
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'user_id': 'str',                                       # injected from auth (required)
                'domain_id': 'str'                                      # injected from auth (required)
            }

        Returns:
            dict:

        """

        query = params.query or {}
        return self.pri_dashboard_mgr.stat_private_dashboards(query)



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.dashboard.manager.public_dashboard_manager import PublicDashboardManager
from spaceone.dashboard.manager.public_folder_manager import PublicFolderManager
from spaceone.dashboard.manager.public_widget_manager import PublicWidgetManager
from spaceone.dashboard.manager.public_data_table_manager import PublicDataTableManager
from spaceone.dashboard.manager.identity_manager import IdentityManager
from spaceone.dashboard.model.public_dashboard.request import *
from spaceone.dashboard.model.public_dashboard.response import *
from spaceone.dashboard.model.public_dashboard.database import PublicDashboard
from spaceone.dashboard.service.public_widget_service import PublicWidgetService
from spaceone.dashboard.error.dashboard import *

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PublicDashboardService(BaseService):
    resource = "PublicDashboard"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pub_dashboard_mgr = PublicDashboardManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def create(
        self, params: PublicDashboardCreateRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Create public dashboard

        Args:
            params (dict): {
                'name': 'str',                  # required
                'description': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'vars_schema': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'dict',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'resource_group': 'str',        # required
                'project_id': 'str',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_info = self.create_dashboard(params.dict())
        return PublicDashboardResponse(**pub_dashboard_info)

    @check_required(["name", "resource_group", "domain_id"])
    def create_dashboard(self, params_dict: dict) -> dict:
        resource_group = params_dict["resource_group"]
        domain_id = params_dict["domain_id"]
        workspace_id = params_dict.get("workspace_id")
        user_projects = params_dict.get("user_projects")

        layouts = params_dict.get("layouts")
        if layouts:
            del params_dict["layouts"]

        if resource_group == "PROJECT":
            if project_id := params_dict.get("project_id"):
                project_info = self.identity_mgr.get_project(project_id)
                params_dict["workspace_id"] = project_info["workspace_id"]
            else:
                raise ERROR_REQUIRED_PARAMETER(key="project_id")

        elif resource_group == "WORKSPACE":
            if workspace_id := params_dict.get("workspace_id"):
                self.identity_mgr.check_workspace(workspace_id, domain_id)
                params_dict["project_id"] = "*"
            else:
                raise ERROR_REQUIRED_PARAMETER(key="workspace_id")
        else:
            params_dict["workspace_id"] = "-"
            params_dict["project_id"] = "-"

        if folder_id := params_dict.get("folder_id"):
            pub_folder_mgr = PublicFolderManager()
            pub_folder_mgr.get_public_folder(
                folder_id,
                domain_id,
                workspace_id,
                user_projects,
                resource_group,
            )

        pub_dashboard_vo = self.pub_dashboard_mgr.create_public_dashboard(params_dict)

        if layouts:
            created_layouts = self._create_layouts(
                layouts,
                pub_dashboard_vo.dashboard_id,
                domain_id,
                workspace_id,
                user_projects,
            )
            self.pub_dashboard_mgr.update_public_dashboard_by_vo(
                {"layouts": created_layouts}, pub_dashboard_vo
            )

        return pub_dashboard_vo.to_dict()

    @staticmethod
    def _create_layouts(
        layouts: list,
        dashboard_id: str,
        domain_id: str,
        workspace_id: str,
        user_projects: list,
    ) -> list:
        pub_widget_svc = PublicWidgetService()
        created_layouts = []
        for layout in layouts:
            widgets = layout.get("widgets", [])
            created_layout = {"name": layout.get("name", ""), "widgets": []}
            for widget in widgets:
                widget["state"] = "ACTIVE"
                widget["dashboard_id"] = dashboard_id
                widget["domain_id"] = domain_id
                widget["workspace_id"] = workspace_id
                widget["user_projects"] = user_projects
                widget["is_bulk"] = True

                widget_info = pub_widget_svc.create_widget(widget)
                created_widget_id = widget_info["widget_id"]

                _LOGGER.debug(f"[create_dashboard] create widget: {created_widget_id}")
                created_layout["widgets"].append(created_widget_id)

            created_layouts.append(created_layout)
        return created_layouts

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def update(
        self, params: PublicDashboardUpdateRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Update public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'description': 'str',
                'name': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'vars_schema': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'list',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if params.folder_id:
            pub_folder_mgr = PublicFolderManager()
            pub_folder_mgr.get_public_folder(
                params.folder_id,
                pub_dashboard_vo.domain_id,
                pub_dashboard_vo.workspace_id,
                params.user_projects,
                pub_dashboard_vo.resource_group,
            )

        pub_dashboard_vo = self.pub_dashboard_mgr.update_public_dashboard_by_vo(
            params.dict(exclude_unset=True), pub_dashboard_vo
        )

        return PublicDashboardResponse(**pub_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def change_folder(
        self, params: PublicDashboardChangeFolderRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Change public dashboard folder

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'folder_id': 'str',             # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }
        Returns:
            PublicDashboardResponse:
        """

        folder_id = params.folder_id

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if folder_id:
            pub_folder_mgr = PublicFolderManager()
            pub_folder_vo = pub_folder_mgr.get_public_folder(
                params.folder_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )

            if pub_dashboard_vo.resource_group != pub_folder_vo.resource_group:
                raise ERROR_INVALID_PARAMETER(
                    key="folder_id",
                    reason="Resource group of dashboard and folder are different.",
                )

        pub_dashboard_vo = self.pub_dashboard_mgr.update_public_dashboard_by_vo(
            params.dict(), pub_dashboard_vo
        )

        pub_dashboard_info = self.unshare_dashboard(
            {
                "dashboard_id": pub_dashboard_vo.dashboard_id,
                "domain_id": pub_dashboard_vo.domain_id,
                "cascade": True,
            }
        )

        if folder_id:
            pub_folder_mgr = PublicFolderManager()
            pub_folder_vo = pub_folder_mgr.get_public_folder(
                params.folder_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )

            if pub_folder_vo.shared:
                pub_dashboard_info = self.share_dashboard(
                    {
                        "dashboard_id": pub_dashboard_vo.dashboard_id,
                        "domain_id": pub_dashboard_vo.domain_id,
                        "scope": pub_folder_vo.scope,
                        "cascade": True,
                    }
                )

        return PublicDashboardResponse(**pub_dashboard_info)

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def share(
        self, params: PublicDashboardShareRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Share public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'scope': 'str',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_info = self.share_dashboard(params.dict())
        return PublicDashboardResponse(**pub_dashboard_info)

    @check_required(["dashboard_id", "domain_id"])
    def share_dashboard(self, params_dict: dict) -> dict:
        dashboard_id = params_dict["dashboard_id"]
        domain_id = params_dict["domain_id"]
        workspace_id = params_dict.get("workspace_id")
        user_projects = params_dict.get("user_projects")
        scope = params_dict.get("scope")
        cascade = params_dict.get("cascade", False)

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            dashboard_id,
            domain_id,
            workspace_id,
            user_projects,
        )

        if pub_dashboard_vo.folder_id and not cascade:
            raise ERROR_NOT_ALLOWED_SHARE(folder_id=pub_dashboard_vo.folder_id)

        updated_params = {
            "shared": True,
        }

        if pub_dashboard_vo.resource_group == "DOMAIN":
            updated_params["workspace_id"] = "*"
            updated_params["project_id"] = "*"
            if scope == "PROJECT":
                updated_params["scope"] = "PROJECT"
            else:
                updated_params["scope"] = "WORKSPACE"
        elif pub_dashboard_vo.resource_group == "WORKSPACE":
            # updated_params["project_id"] = "*"
            updated_params["scope"] = "PROJECT"
        elif pub_dashboard_vo.resource_group == "PROJECT":
            raise ERROR_PERMISSION_DENIED()

        pub_dashboard_vo = self.pub_dashboard_mgr.update_public_dashboard_by_vo(
            updated_params, pub_dashboard_vo
        )

        # if pub_dashboard_vo.resource_group in ["DOMAIN", "WORKSPACE"]:
        if pub_dashboard_vo.resource_group == "DOMAIN":
            # Cascade update for widgets
            pub_widget_mgr = PublicWidgetManager()
            pub_widget_vos = pub_widget_mgr.filter_public_widgets(
                dashboard_id=pub_dashboard_vo.dashboard_id,
                domain_id=pub_dashboard_vo.domain_id,
            )
            for pub_widget_vo in pub_widget_vos:
                pub_widget_mgr.update_public_widget_by_vo(updated_params, pub_widget_vo)

            # Cascade update for data tables
            pub_data_table_mgr = PublicDataTableManager()
            pub_data_table_vos = pub_data_table_mgr.filter_public_data_tables(
                dashboard_id=pub_dashboard_vo.dashboard_id,
                domain_id=pub_dashboard_vo.domain_id,
            )
            for pub_data_table_vo in pub_data_table_vos:
                pub_data_table_mgr.update_public_data_table_by_vo(
                    updated_params, pub_data_table_vo
                )

        return pub_dashboard_vo.to_dict()

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def unshare(
        self, params: PublicDashboardUnshareRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Unshare public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_info = self.unshare_dashboard(params.dict())
        return PublicDashboardResponse(**pub_dashboard_info)

    @check_required(["dashboard_id", "domain_id"])
    def unshare_dashboard(self, params_dict: dict) -> dict:
        dashboard_id = params_dict["dashboard_id"]
        domain_id = params_dict["domain_id"]
        workspace_id = params_dict.get("workspace_id")
        user_projects = params_dict.get("user_projects")
        cascade = params_dict.get("cascade", False)

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            dashboard_id,
            domain_id,
            workspace_id,
            user_projects,
        )

        if pub_dashboard_vo.folder_id and not cascade:
            raise ERROR_NOT_ALLOWED_UNSHARE(folder_id=pub_dashboard_vo.folder_id)

        updated_params = {
            "shared": False,
        }

        if pub_dashboard_vo.resource_group == "DOMAIN":
            updated_params["workspace_id"] = "-"
            updated_params["project_id"] = "-"
            updated_params["scope"] = None
        elif pub_dashboard_vo.resource_group == "WORKSPACE":
            # updated_params["project_id"] = "-"
            updated_params["scope"] = None
        elif pub_dashboard_vo.resource_group == "PROJECT":
            raise ERROR_PERMISSION_DENIED()

        pub_dashboard_vo = self.pub_dashboard_mgr.update_public_dashboard_by_vo(
            updated_params, pub_dashboard_vo
        )

        # if pub_dashboard_vo.resource_group in ["DOMAIN", "WORKSPACE"]:
        if pub_dashboard_vo.resource_group == "DOMAIN":
            # Cascade update for widgets
            pub_widget_mgr = PublicWidgetManager()
            pub_widget_vos = pub_widget_mgr.filter_public_widgets(
                dashboard_id=pub_dashboard_vo.dashboard_id,
                domain_id=pub_dashboard_vo.domain_id,
            )
            for pub_widget_vo in pub_widget_vos:
                pub_widget_mgr.update_public_widget_by_vo(updated_params, pub_widget_vo)

            # Cascade update for data tables
            pub_data_table_mgr = PublicDataTableManager()
            pub_data_table_vos = pub_data_table_mgr.filter_public_data_tables(
                dashboard_id=pub_dashboard_vo.dashboard_id,
                domain_id=pub_dashboard_vo.domain_id,
            )
            for pub_data_table_vo in pub_data_table_vos:
                pub_data_table_mgr.update_public_data_table_by_vo(
                    updated_params, pub_data_table_vo
                )

        return pub_dashboard_vo.to_dict()

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def delete(self, params: PublicDashboardDeleteRequest) -> None:
        """Delete public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        self.pub_dashboard_mgr.delete_public_dashboard_by_vo(pub_dashboard_vo)

    @transaction(
        permission="dashboard:PublicDashboard.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "user_projects", "*")
    @convert_model
    def get(
        self, params: PublicDashboardGetRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Get public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        return PublicDashboardResponse(**pub_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDashboard.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "user_projects", "*")
    @append_query_filter(
        [
            "dashboard_id",
            "name",
            "domain_id",
            "workspace_id",
            "project_id",
            "folder_id",
            "user_projects",
        ]
    )
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def list(
        self, params: PublicDashboardSearchQueryRequest
    ) -> Union[PublicDashboardsResponse, dict]:
        """List public dashboards

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',
                'name': 'str',
                'folder_id': 'str',
                'project_id': 'str',
                'workspace_id': 'str',                          # injected from auth
                'domain_id': 'str',                             # injected from auth (required)
                'user_projects': 'list',                        # injected from auth
            }

        Returns:
            PublicDashboardsResponse:
        """

        query = params.query or {}
        pub_dashboard_vos, total_count = self.pub_dashboard_mgr.list_public_dashboards(
            query
        )
        pub_dashboards_info = [
            pub_dashboard_vo.to_dict() for pub_dashboard_vo in pub_dashboard_vos
        ]
        return PublicDashboardsResponse(
            results=pub_dashboards_info, total_count=total_count
        )

    @transaction(
        permission="dashboard:PublicDashboard.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "user_projects", "*")
    @append_query_filter(["domain_id", "workspace_id", "user_projects"])
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def stat(self, params: PublicDashboardStatQueryRequest) -> dict:
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'workspace_id': 'str',                                  # injected from auth
                'domain_id': 'str'                                      # injected from auth (required)
                'user_projects': 'list',                                # injected from auth
            }

        Returns:
            dict:

        """

        query = params.query or {}
        return self.pub_dashboard_mgr.stat_public_dashboards(query)



import logging
from typing import Tuple
from mongoengine import QuerySet

from spaceone.core.manager import BaseManager
from spaceone.dashboard.model.public_widget.database import PublicWidget
from spaceone.dashboard.manager.public_data_table_manager import PublicDataTableManager

_LOGGER = logging.getLogger(__name__)


class PublicWidgetManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget_model = PublicWidget

    def create_public_widget(self, params: dict) -> PublicWidget:
        def _rollback(vo: PublicWidget) -> None:
            _LOGGER.info(
                f"[create_public_widget._rollback] "
                f"Delete vo : {vo.name} "
                f"({vo.widget_id})"
            )
            vo.delete()

        widget_vo: PublicWidget = self.widget_model.create(params)
        self.transaction.add_rollback(_rollback, widget_vo)

        return widget_vo

    def update_public_widget_by_vo(
        self, params: dict, widget_vo: PublicWidget
    ) -> PublicWidget:
        def _rollback(old_data: dict) -> None:
            _LOGGER.info(
                f"[update_public_widget_by_vo._rollback] Revert Data : "
                f'{old_data["widget_id"]}'
            )
            widget_vo.update(old_data)

        self.transaction.add_rollback(_rollback, widget_vo.to_dict())
        return widget_vo.update(params)

    @staticmethod
    def delete_public_widget_by_vo(widget_vo: PublicWidget) -> None:
        # Delete child data tables
        pub_data_table_mgr = PublicDataTableManager()
        pub_data_table_vos = pub_data_table_mgr.filter_public_data_tables(
            widget_id=widget_vo.widget_id,
            domain_id=widget_vo.domain_id,
        )
        pub_data_table_vos.delete()

        widget_vo.delete()

    def get_public_widget(
        self,
        widget_id: str,
        domain_id: str,
        workspace_id: str = None,
        user_projects=None,
    ) -> PublicWidget:
        conditions = {
            "widget_id": widget_id,
            "domain_id": domain_id,
        }

        if workspace_id:
            conditions["workspace_id"] = workspace_id

        if user_projects:
            conditions["project_id"] = user_projects

        return self.widget_model.get(**conditions)

    def filter_public_widgets(self, **conditions) -> QuerySet:
        return self.widget_model.filter(**conditions)

    def list_public_widgets(self, query: dict) -> Tuple[QuerySet, int]:
        return self.widget_model.query(**query)

    def stat_public_widgets(self, query: dict) -> dict:
        return self.widget_model.stat(**query)



import logging
from typing import Tuple
from mongoengine import QuerySet

from spaceone.core.manager import BaseManager
from spaceone.dashboard.model.private_dashboard.database import PrivateDashboard
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)
from spaceone.dashboard.manager.private_widget_manager import PrivateWidgetManager

_LOGGER = logging.getLogger(__name__)


class PrivateDashboardManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dashboard_model = PrivateDashboard

    def create_private_dashboard(self, params: dict) -> PrivateDashboard:
        def _rollback(vo: PrivateDashboard) -> None:
            _LOGGER.info(
                f"[create_private_dashboard._rollback] "
                f"Delete vo : {vo.name} "
                f"({vo.dashboard_id})"
            )
            vo.delete()

        dashboard_vo: PrivateDashboard = self.dashboard_model.create(params)
        self.transaction.add_rollback(_rollback, dashboard_vo)

        return dashboard_vo

    def update_private_dashboard_by_vo(
        self, params: dict, dashboard_vo: PrivateDashboard
    ) -> PrivateDashboard:
        def _rollback(old_data: dict) -> None:
            _LOGGER.info(
                f"[update_private_dashboard_by_vo._rollback] Revert Data : "
                f'{old_data["dashboard_id"]}'
            )
            dashboard_vo.update(old_data)

        self.transaction.add_rollback(_rollback, dashboard_vo.to_dict())
        return dashboard_vo.update(params)

    @staticmethod
    def delete_private_dashboard_by_vo(dashboard_vo: PrivateDashboard) -> None:
        # Delete child widgets
        pri_widget_mgr = PrivateWidgetManager()
        pri_widget_vos = pri_widget_mgr.filter_private_widgets(
            dashboard_id=dashboard_vo.dashboard_id,
            domain_id=dashboard_vo.domain_id,
        )
        pri_widget_vos.delete()

        # Delete child data tables
        pri_data_table_mgr = PrivateDataTableManager()
        pri_data_table_vos = pri_data_table_mgr.filter_private_data_tables(
            dashboard_id=dashboard_vo.dashboard_id,
            domain_id=dashboard_vo.domain_id,
        )
        pri_data_table_vos.delete()

        dashboard_vo.delete()

    def get_private_dashboard(
        self,
        dashboard_id: str,
        domain_id: str,
        user_id: str,
    ) -> PrivateDashboard:
        return self.dashboard_model.get(
            dashboard_id=dashboard_id, domain_id=domain_id, user_id=user_id
        )

    def filter_private_dashboards(self, **conditions) -> QuerySet:
        return self.dashboard_model.filter(**conditions)

    def list_private_dashboards(self, query: dict) -> Tuple[QuerySet, int]:
        return self.dashboard_model.query(**query)

    def stat_private_dashboards(self, query: dict) -> dict:
        return self.dashboard_model.stat(**query)



import logging
from typing import Tuple
from mongoengine import QuerySet

from spaceone.core.manager import BaseManager
from spaceone.dashboard.model.private_data_table.database import PrivateDataTable

_LOGGER = logging.getLogger(__name__)


class PrivateDataTableManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_table_model = PrivateDataTable

    def create_private_data_table(self, params: dict) -> PrivateDataTable:
        def _rollback(vo: PrivateDataTable) -> None:
            _LOGGER.info(
                f"[create_private_data_table._rollback] "
                f"Delete vo : {vo.name} "
                f"({vo.data_table_id})"
            )
            vo.delete()

        data_table_vo: PrivateDataTable = self.data_table_model.create(params)
        self.transaction.add_rollback(_rollback, data_table_vo)

        return data_table_vo

    def update_private_data_table_by_vo(
        self, params: dict, data_table_vo: PrivateDataTable
    ) -> PrivateDataTable:
        def _rollback(old_data: dict) -> None:
            _LOGGER.info(
                f"[update_private_data_table_by_vo._rollback] Revert Data : "
                f'{old_data["data_table_id"]}'
            )
            data_table_vo.update(old_data)

        self.transaction.add_rollback(_rollback, data_table_vo.to_dict())
        return data_table_vo.update(params)

    @staticmethod
    def delete_private_data_table_by_vo(data_table_vo: PrivateDataTable) -> None:
        data_table_vo.delete()

    def get_private_data_table(
        self,
        data_table_id: str,
        domain_id: str,
        user_id: str = None,
    ) -> PrivateDataTable:
        conditions = {
            "data_table_id": data_table_id,
            "domain_id": domain_id,
        }

        if user_id:
            conditions.update({"user_id": user_id})

        return self.data_table_model.get(**conditions)

    def filter_private_data_tables(self, **conditions) -> QuerySet:
        return self.data_table_model.filter(**conditions)

    def list_private_data_tables(self, query: dict) -> Tuple[QuerySet, int]:
        return self.data_table_model.query(**query)

    def stat_private_data_tables(self, query: dict) -> dict:
        return self.data_table_model.stat(**query)



from spaceone.api.dashboard.v1 import public_folder_pb2, public_folder_pb2_grpc
from spaceone.core.pygrpc import BaseAPI
from spaceone.dashboard.service.public_folder_service import PublicFolderService


class PublicFolder(BaseAPI, public_folder_pb2_grpc.PublicFolderServicer):
    pb2 = public_folder_pb2
    pb2_grpc = public_folder_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.create(params)
        return self.dict_to_message(response)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.update(params)
        return self.dict_to_message(response)

    def share(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.share(params)
        return self.dict_to_message(response)

    def unshare(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.unshare(params)
        return self.dict_to_message(response)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        pub_folder_svc.delete(params)
        return self.empty()

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.get(params)
        return self.dict_to_message(response)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.list(params)
        return self.dict_to_message(response)

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)
        pub_folder_svc = PublicFolderService(metadata)
        response: dict = pub_folder_svc.stat(params)
        return self.dict_to_message(response)



import logging
from typing import List, Union, Tuple

import numpy as np
import pandas as pd

from spaceone.dashboard.error.data_table import (
    ERROR_INVALID_PARAMETER,
    ERROR_NOT_SUPPORTED_OPERATOR,
    ERROR_REQUIRED_PARAMETER,
    ERROR_DUPLICATED_DATA_FIELDS,
    ERROR_INVALID_PARAMETER_TYPE,
    ERROR_DUPLICATED_FIELD_NAME,
    ERROR_NOT_ALLOWED_DATA_FIELD,
)
from spaceone.dashboard.manager.data_table_manager import DataTableManager
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.public_data_table_manager import PublicDataTableManager
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)
from spaceone.dashboard.model.public_data_table.database import PublicDataTable
from spaceone.dashboard.model.private_data_table.database import PrivateDataTable

_LOGGER = logging.getLogger(__name__)


class DataTransformationManager(DataTableManager):
    def __init__(
        self,
        data_table_type: str,
        operator: str,
        options: dict,
        widget_id: str,
        domain_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if operator not in [
            "JOIN",
            "CONCAT",
            "AGGREGATE",
            "QUERY",
            "EVAL",
            "PIVOT",
            "ADD_LABELS",
            "VALUE_MAPPING",
        ]:
            raise ERROR_NOT_SUPPORTED_OPERATOR(operator=operator)

        self.data_table_type = data_table_type
        self.operator = operator
        self.options = options
        self.widget_id = widget_id
        self.domain_id = domain_id
        self.data_table_vos = self._get_data_table_from_options(operator, options)
        self.data_keys = []
        self.label_keys = []
        self.total_series = None

    def get_data_and_labels_info(self) -> Tuple[dict, dict]:
        data_info = {}
        labels_info = {}
        for data_table_vo in self.data_table_vos:
            for key, value in data_table_vo.data_info.items():
                if key not in data_info and key in self.data_keys:
                    data_info[key] = value

            for key, value in data_table_vo.labels_info.items():
                if key not in labels_info and key in self.label_keys:
                    labels_info[key] = value

        for key in self.data_keys:
            if key not in data_info:
                data_info[key] = {}

        for key in self.label_keys:
            if key not in labels_info:
                labels_info[key] = {}

        return data_info, labels_info

    def load(
        self,
        granularity: str = "MONTHLY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> pd.DataFrame:
        try:
            if self.operator == "JOIN":
                self.join_data_tables(granularity, start, end, vars)
            elif self.operator == "CONCAT":
                self.concat_data_tables(granularity, start, end, vars)
            elif self.operator == "AGGREGATE":
                self.aggregate_data_table(granularity, start, end, vars)
            elif self.operator == "QUERY":
                self.query_data_table(granularity, start, end, vars)
            elif self.operator == "EVAL":
                self.evaluate_data_table(granularity, start, end, vars)
            elif self.operator == "PIVOT":
                self.pivot_data_table(granularity, start, end, vars)
            elif self.operator == "ADD_LABELS":
                self.add_labels_data_table(granularity, start, end, vars)
            elif self.operator == "VALUE_MAPPING":
                self.value_mapping_data_table(granularity, start, end, vars)

            self.state = "AVAILABLE"
            self.error_message = None

        except Exception as e:
            self.state = "UNAVAILABLE"
            self.error_message = e.message if hasattr(e, "message") else str(e)
            _LOGGER.error(f"[load] {self.operator} operation error: {e}", exc_info=True)

        return self.df

    def join_data_tables(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        how = self.options.get("how", "left")
        left_keys = self.options.get("left_keys")
        right_keys = self.options.get("right_keys")

        self._validate_options(how, left_keys, right_keys)

        origin_vo = self.data_table_vos[0]
        other_vo = self.data_table_vos[1]
        origin_df = self._get_data_table(origin_vo, granularity, start, end, vars)
        other_df = self._get_data_table(other_vo, granularity, start, end, vars)

        self._validate_join_keys(left_keys, right_keys, origin_vo, other_vo)

        self._set_data_keys(origin_vo, other_vo)

        merged_df = self._merge_data_frames(
            origin_df, other_df, how, left_keys, right_keys
        )

        merged_df = self._rename_duplicated_columns(
            merged_df, origin_vo.name, other_vo.name
        )

        self.label_keys = list(set(merged_df.columns) - set(self.data_keys))
        self.df = merged_df

    def concat_data_tables(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        origin_vo = self.data_table_vos[0]
        other_vo = self.data_table_vos[1]
        origin_data_keys = set(origin_vo.data_info.keys())
        other_data_keys = set(other_vo.data_info.keys())

        self.data_keys = list(origin_data_keys | other_data_keys)
        origin_label_keys = set(origin_vo.labels_info.keys())
        other_label_keys = set(other_vo.labels_info.keys())
        self.label_keys = list(origin_label_keys | other_label_keys)

        fill_na = {}
        for key in self.data_keys:
            fill_na[key] = 0

        for key in self.label_keys:
            fill_na[key] = ""

        origin_df = self._get_data_table(origin_vo, granularity, start, end, vars)
        other_df = self._get_data_table(other_vo, granularity, start, end, vars)

        merged_df = pd.concat([origin_df, other_df], ignore_index=True)
        merged_df = merged_df.fillna(value=fill_na)
        self.df = merged_df

    def aggregate_data_table(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        function = dict(self.options.get("function", {}))
        if function == {}:
            raise ERROR_REQUIRED_PARAMETER(key="options.AGGREGATE.function")

        for key, operator in function.items():
            if operator not in ["sum", "mean", "max", "min", "count"]:
                raise ERROR_INVALID_PARAMETER(
                    key="options.AGGREGATE.function",
                    reason=f"Invalid function type: {operator}",
                )

        group_by = list(self.options.get("group_by", []))

        origin_vo = self.data_table_vos[0]
        self.data_keys = list(function.keys())
        self.label_keys = group_by

        df = self._get_data_table(origin_vo, granularity, start, end, vars)

        for key in group_by:
            if key not in df.columns:
                raise ERROR_INVALID_PARAMETER(
                    key="options.AGGREGATE.group_by", reason=f"Invalid key: {key}"
                )

        for key in function.keys():
            if key not in df.columns:
                raise ERROR_INVALID_PARAMETER(
                    key="options.AGGREGATE.function", reason=f"Invalid key: {key}"
                )

        if len(group_by) > 0:
            self.df = df.groupby(group_by).agg(function).reset_index()
        else:
            self.df = df.agg(function).to_frame().T

    def query_data_table(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        conditions = self.options.get("conditions", [])

        origin_vo = self.data_table_vos[0]
        self.data_keys = list(origin_vo.data_info.keys())
        self.label_keys = list(origin_vo.labels_info.keys())

        df = self._get_data_table(origin_vo, granularity, start, end, vars)

        for condition in conditions:
            if self.is_jinja_expression(condition):
                condition, gv_type_map = self.change_global_variables(condition, vars)
                condition = self.remove_jinja_braces(condition)
                condition = self.change_expression_data_type(condition, gv_type_map)
                condition = self.change_space_variable(condition)

            try:
                df = df.query(condition)
            except Exception as e:
                _LOGGER.error(f"[query_data_table] query error: {e}")
                raise ERROR_INVALID_PARAMETER(
                    key="options.QUERY.conditions", reason=condition
                )

        self.df = df

    def evaluate_data_table(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        expressions = self.options.get("expressions", [])

        origin_vo = self.data_table_vos[0]
        self.data_keys = list(origin_vo.data_info.keys())
        self.label_keys = list(origin_vo.labels_info.keys())

        df = self._get_data_table(origin_vo, granularity, start, end, vars)

        for expression in expressions:
            if isinstance(expression, dict):
                name = expression.get("name")
                field_type = expression.get("field_type", "DATA")
                condition = expression.get("condition")
                value_expression = expression.get("expression")

                if name is None:
                    raise ERROR_REQUIRED_PARAMETER(key="options.EVAL.expressions.name")

                if name in self.data_keys:
                    raise ERROR_DUPLICATED_FIELD_NAME(field=name, fields=self.data_keys)

                if value_expression is None:
                    raise ERROR_REQUIRED_PARAMETER(
                        key="options.EVAL.expressions.expression"
                    )

                if self.is_jinja_expression(condition):
                    condition, gv_type_map = self.change_global_variables(
                        condition, vars
                    )
                    condition = self.remove_jinja_braces(condition)
                    condition = self.change_expression_data_type(condition, gv_type_map)
                    condition = self.change_space_variable(condition)

                if self.is_jinja_expression(value_expression):
                    value_expression, gv_type_map = self.change_global_variables(
                        value_expression, vars
                    )
                    value_expression = self.remove_jinja_braces(value_expression)
                    value_expression = self.change_expression_data_type(
                        value_expression, gv_type_map
                    )
                    value_expression = self.change_space_variable(value_expression)

                template_vars = {}
                for key in self.data_keys:
                    template_vars[key] = f"`{key}`"

                for key in self.label_keys:
                    template_vars[key] = f"`{key}`"

                if field_type not in ["DATA", "LABEL"]:
                    raise ERROR_INVALID_PARAMETER(
                        key="options.EVAL.expressions.field_type",
                        reason=f"Invalid field type: {field_type}",
                    )

                if isinstance(value_expression, str):
                    try:
                        value_expression = value_expression.format(**template_vars)
                    except KeyError as e:
                        value_expression = value_expression.replace("{", "{{").replace(
                            "}", "}}"
                        )
                        value_expression = value_expression.format(**template_vars)
                    except Exception as e:
                        raise ERROR_INVALID_PARAMETER(
                            key="options.EVAL.expressions.expression",
                            reason=f"Invalid expression: (template_var={e})",
                        )

                    if "@" in value_expression:
                        raise ERROR_INVALID_PARAMETER(
                            key="options.EVAL.expressions",
                            reason="It should not have '@' symbol.",
                        )

                try:
                    merged_expr = f"`{name}` = {value_expression}"

                    if field_type == "LABEL":
                        self.label_keys = list(set(self.label_keys) | {name})
                    else:
                        self.data_keys = list(set(self.data_keys) | {name})

                    if name in df.columns:
                        last_key = name
                    else:
                        last_key = df.eval(merged_expr).columns[-1:][0]

                    if not df.empty:
                        if condition:
                            matched_index = df.query(condition).index
                            if " " in name and name in df.columns:
                                temp_key = name.replace(" ", "_")
                                df.rename(columns={name: temp_key}, inplace=True)
                                merged_expr = f"`{temp_key}` = {value_expression}"
                                df.loc[matched_index, temp_key] = df.eval(
                                    merged_expr
                                ).loc[matched_index, temp_key]
                                df.rename(columns={temp_key: name}, inplace=True)
                            else:
                                df.loc[matched_index, last_key] = df.eval(
                                    merged_expr
                                ).loc[matched_index, last_key]

                        else:
                            df.eval(merged_expr, inplace=True)

                    if last_key.startswith("BACKTICK_QUOTED_STRING"):
                        df.rename(columns={last_key: name}, inplace=True)

                    df.replace([np.inf, -np.inf], 0, inplace=True)

                except Exception as e:
                    _LOGGER.error(f"[evaluate_data_table] eval error: {e}")
                    raise ERROR_INVALID_PARAMETER(
                        key="options.EVAL.expressions", reason=expression
                    )

            elif isinstance(expression, str):
                if "@" in expression:
                    raise ERROR_INVALID_PARAMETER(
                        key="options.EVAL.expressions",
                        reason="It should not have '@' symbol.",
                    )

                try:
                    name, value_expression = expression.split("=", 1)
                    name = name.replace("`", "").strip()
                    value_expression = value_expression.strip()
                    expression = f"`{name}` = {value_expression}"

                    self.data_keys = list(set(self.data_keys) | {name})

                    df.eval(expression, inplace=True)
                    last_key = df.columns[-1:][0]
                    if last_key.startswith("BACKTICK_QUOTED_STRING"):
                        df.rename(columns={last_key: name}, inplace=True)
                    df.replace([np.inf, -np.inf], 0, inplace=True)

                except Exception as e:
                    _LOGGER.error(f"[evaluate_data_table] eval error: {e}")
                    raise ERROR_INVALID_PARAMETER(
                        key="options.EVAL.expressions", reason=expression
                    )
            else:
                raise ERROR_INVALID_PARAMETER_TYPE(
                    key="options.EVAL.expressions", type=type(expression)
                )

        self.df = df

    def pivot_data_table(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        origin_vo = self.data_table_vos[0]
        field_options = self.options["fields"]
        labels, column, data = (
            field_options.get("labels"),
            field_options["column"],
            field_options["data"],
        )
        function = self.options.get("function", "sum")

        raw_df = self._get_data_table(origin_vo, granularity, start, end, vars)

        if raw_df.empty:
            self.df = raw_df
        else:
            self._check_columns(raw_df, labels, column, data)
            fill_value = self._set_fill_value_from_df(raw_df, data)

            pivot_table = self._create_pivot_table(
                raw_df, labels, column, data, function, fill_value
            )

            pivot_table = self._sort_and_filter_pivot_table(pivot_table)

            self.df = pivot_table

    def add_labels_data_table(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        data_table_vo = self.data_table_vos[0]
        label_keys = list(data_table_vo.labels_info.keys())
        data_keys = list(data_table_vo.data_info.keys())
        labels = self.options.get("labels")

        df = self._get_data_table(data_table_vo, granularity, start, end, vars)

        self.validate_labels(labels, df)

        self.add_labels_to_dataframe(df, labels, label_keys, data_keys)

        df = df.reindex(columns=label_keys + data_keys)

        self.label_keys = label_keys
        self.data_keys = data_keys

        self.df = df

    def value_mapping_data_table(
        self,
        granularity: str = "MONTHLY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        data_table_vo = self.data_table_vos[0]
        df = self._get_data_table(data_table_vo, granularity, start, end, vars)

        self.label_keys = list(data_table_vo.labels_info.keys())
        self.data_keys = list(data_table_vo.data_info.keys())

        name = self.options["name"]
        if name in self.data_keys:
            raise ERROR_NOT_ALLOWED_DATA_FIELD(
                name=name, data_fields=list(self.data_keys)
            )

        field_type = self.options.get("field_type", "LABEL")

        filtered_df = self.filter_data(df, vars)
        filtered_df = self.apply_cases(filtered_df)

        df.loc[filtered_df.index, name] = filtered_df[name]
        self.handle_unfiltered_data(df, filtered_df, name, field_type)

        self.df = df

    def _get_data_table(
        self,
        data_table_vo: Union[PublicDataTable, PrivateDataTable],
        granularity: str,
        start: str,
        end: str,
        vars: dict,
    ) -> pd.DataFrame:
        if data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                self.data_table_type,
                data_table_vo.source_type,
                data_table_vo.options,
                data_table_vo.widget_id,
                data_table_vo.domain_id,
            )
            return ds_mgr.load(granularity, start, end, vars)
        else:
            operator = data_table_vo.operator
            options = data_table_vo.options.get(operator, {})
            ds_mgr = DataTransformationManager(
                self.data_table_type,
                data_table_vo.operator,
                options,
                data_table_vo.widget_id,
                data_table_vo.domain_id,
            )

            return ds_mgr.load(granularity, start, end, vars)

    def _get_data_table_from_options(
        self,
        operator: str,
        options: dict,
    ) -> List[Union[PublicDataTable, PrivateDataTable]]:
        parent_dt_vos = []

        if self.data_table_type == "PUBLIC":
            data_table_mgr = PublicDataTableManager()
        else:
            data_table_mgr = PrivateDataTableManager()

        if operator in ["JOIN", "CONCAT"]:
            if data_tables := options.get("data_tables"):
                if len(data_tables) != 2:
                    raise ERROR_INVALID_PARAMETER(
                        key="options.data_tables",
                        reason="It should have 2 data tables.",
                    )

                for data_table_id in data_tables:
                    if self.data_table_type == "PUBLIC":
                        parent_dt_vo = data_table_mgr.get_public_data_table(
                            data_table_id,
                            self.domain_id,
                        )
                    else:
                        parent_dt_vo = data_table_mgr.get_private_data_table(
                            data_table_id,
                            self.domain_id,
                        )

                    if parent_dt_vo.widget_id != self.widget_id:
                        raise ERROR_INVALID_PARAMETER(
                            key="options.data_tables",
                            reason="It should have same widget_id.",
                        )
                    parent_dt_vos.append(parent_dt_vo)

            else:
                raise ERROR_REQUIRED_PARAMETER(key=f"options.{operator}.data_tables")
        else:
            if data_table_id := options.get("data_table_id"):
                if self.data_table_type == "PUBLIC":
                    parent_dt_vo = data_table_mgr.get_public_data_table(
                        data_table_id,
                        self.domain_id,
                    )
                else:
                    parent_dt_vo = data_table_mgr.get_private_data_table(
                        data_table_id,
                        self.domain_id,
                    )

                if parent_dt_vo.widget_id != self.widget_id:
                    raise ERROR_INVALID_PARAMETER(
                        key="options.data_table_id",
                        reason="It should have same widget_id.",
                    )
                parent_dt_vos.append(parent_dt_vo)
            else:
                raise ERROR_REQUIRED_PARAMETER(key="options.data_table_id")

        return parent_dt_vos

    def _set_data_keys(
        self,
        origin_vo: Union[PublicDataTable, PrivateDataTable],
        other_vo: Union[PublicDataTable, PrivateDataTable],
    ) -> None:
        origin_data_keys = set(origin_vo.data_info.keys())
        other_data_keys = set(other_vo.data_info.keys())

        duplicate_keys = list(origin_data_keys & other_data_keys)
        if duplicate_keys:
            raise ERROR_DUPLICATED_DATA_FIELDS(field=", ".join(duplicate_keys))

        self.data_keys = list(origin_data_keys | other_data_keys)

    @staticmethod
    def _validate_options(how: str, left_keys: list, right_keys: list) -> None:
        if how not in ["left", "right", "inner", "outer"]:
            raise ERROR_INVALID_PARAMETER(
                key="options.JOIN.how", reason="Invalid join type"
            )
        if not left_keys or not right_keys:
            missing_key = "left_keys" if not left_keys else "right_keys"
            raise ERROR_REQUIRED_PARAMETER(key=f"options.JOIN.{missing_key}")

    @staticmethod
    def _validate_join_keys(
        left_keys: list,
        right_keys: list,
        origin_vo: Union[PublicDataTable, PrivateDataTable],
        other_vo: Union[PublicDataTable, PrivateDataTable],
    ) -> None:
        origin_label_keys = set(origin_vo.labels_info.keys())
        other_label_keys = set(other_vo.labels_info.keys())

        if len(left_keys) != len(right_keys):
            raise ERROR_INVALID_PARAMETER(
                key="options.JOIN",
                reason=f"left_keys and right_keys should have the same length. "
                f"left_keys={list(origin_label_keys)}, right_keys={list(other_label_keys)}",
            )

        for key in left_keys:
            if key not in origin_label_keys:
                raise ERROR_INVALID_PARAMETER(
                    key="options.JOIN.left_keys",
                    reason=f"Invalid key: {key}, table keys={list(origin_label_keys)}",
                )
        for key in right_keys:
            if key not in other_label_keys:
                raise ERROR_INVALID_PARAMETER(
                    key="options.JOIN.right_keys",
                    reason=f"Invalid key: {key}, table keys={list(other_label_keys)}",
                )

    @staticmethod
    def _create_rename_columns_from_join_keys(left_keys, right_keys, how):
        multi_keys = zip(left_keys, right_keys)
        rename_columns = {}
        if how in ["left", "inner", "outer"]:
            for left_key, right_key in multi_keys:
                rename_columns[right_key] = left_key
        elif how == "right":
            for left_key, right_key in multi_keys:
                rename_columns[left_key] = right_key

        return rename_columns

    def _merge_data_frames(
        self,
        origin_df: pd.DataFrame,
        other_df: pd.DataFrame,
        how: str,
        left_keys: list,
        right_keys: list,
    ) -> pd.DataFrame:
        rename_columns = self._create_rename_columns_from_join_keys(
            left_keys, right_keys, how
        )
        if how in ["left", "inner", "outer"]:
            other_df = other_df.rename(columns=rename_columns)
        elif how == "right":
            origin_df = origin_df.rename(columns=rename_columns)

        if origin_df.empty:
            origin_df = pd.DataFrame(columns=left_keys)

        if other_df.empty:
            other_df = pd.DataFrame(columns=right_keys)

        join_keys = left_keys if how in ["left", "inner", "outer"] else right_keys
        merged_df = origin_df.merge(
            other_df, left_on=join_keys, right_on=join_keys, how=how
        )

        label_keys = list(set(merged_df.columns) - set(self.data_keys))
        fill_na = {key: 0 for key in self.data_keys}
        fill_na.update({key: "" for key in label_keys})
        merged_df = merged_df.fillna(value=fill_na)

        return merged_df

    @staticmethod
    def _rename_duplicated_columns(
        merged_df: pd.DataFrame, origin_vo_name: str, other_vo_name: str
    ) -> pd.DataFrame:
        merged_rename_columns = {}
        for column in merged_df.columns:
            if column.endswith("_x"):
                column_name, _ = column.split("_")
                merged_rename_columns[column] = f"{column_name}({origin_vo_name})"
            elif column.endswith("_y"):
                column_name, _ = column.split("_")
                merged_rename_columns[column] = f"{column_name}({other_vo_name})"

        return merged_df.rename(columns=merged_rename_columns)

    @staticmethod
    def _check_columns(
        df: pd.DataFrame, label_fields: list, column_field: str, data_field: str
    ) -> None:
        df_columns = set(df.columns)
        if label_fields:
            for label_field in label_fields:
                if label_field not in df_columns:
                    raise ERROR_INVALID_PARAMETER(
                        key=f"options.PIVOT.labels",
                        reason=f"Invalid key: {label_field}, columns={list(df_columns)}",
                    )

        if column_field not in df_columns:
            raise ERROR_INVALID_PARAMETER(
                key=f"options.PIVOT.column",
                reason=f"Invalid key: {column_field}, columns={list(df_columns)}",
            )

        if data_field not in df_columns:
            raise ERROR_INVALID_PARAMETER(
                key=f"options.PIVOT.data",
                reason=f"Invalid key: {data_field}, columns={list(df_columns)}",
            )

    @staticmethod
    def _set_fill_value_from_df(df: pd.DataFrame, data_field: str) -> Union[int, str]:
        if df[data_field].dtype == "object":
            return ""
        return 0

    @staticmethod
    def _set_new_column_names(pivot_table: pd.DataFrame) -> pd.DataFrame:
        if pivot_table.columns[0] == "index":
            pivot_table = pivot_table.iloc[:, 1:]
        else:
            new_columns = [
                lower_col if lower_col else upper_col
                for upper_col, lower_col in pivot_table.columns
            ]
            pivot_table.columns = new_columns
        return pivot_table

    def _set_keys(self, columns: list) -> None:
        if columns[0] == "index":
            self.label_keys = []
            self.data_keys = [col for col in columns if col != "index"]
        else:
            self.label_keys = [
                upper_col for upper_col, lower_col in columns if not lower_col
            ]
            self.data_keys = [
                lower_col for upper_col, lower_col in columns if lower_col
            ]

    @staticmethod
    def _validate_select_fields(
        select_fields: list,
        column_fields: list,
    ) -> None:
        for select_field in select_fields:
            if select_field not in column_fields:
                raise ERROR_INVALID_PARAMETER(
                    key="options.PIVOT.select",
                    reason=f"Invalid key: {select_field}, columns={column_fields}",
                )

    @staticmethod
    def _validate_function(function: str) -> None:
        if function not in ["sum", "mean", "max", "min"]:
            raise ERROR_INVALID_PARAMETER(
                key="options.PIVOT.function",
                reason=f"Invalid function type: {function}",
            )

    def _create_pivot_table(
        self,
        raw_df: pd.DataFrame,
        labels: list,
        column: str,
        data: str,
        function: str,
        fill_value: Union[int, str],
    ) -> pd.DataFrame:
        try:
            self._validate_function(function)
            pivot_table = pd.pivot_table(
                raw_df,
                values=[data],
                index=labels,
                columns=[column],
                aggfunc=function,
                fill_value=fill_value,
            )
            pivot_table.reset_index(inplace=True)
            self._set_keys(list(pivot_table.columns))
            return self._set_new_column_names(pivot_table)
        except Exception as e:
            _LOGGER.error(f"[pivot_data_table] pivot error: {e}")
            raise ERROR_INVALID_PARAMETER(key="options.PIVOT", reason=str(e))

    def _sort_and_filter_pivot_table(self, pivot_table: pd.DataFrame) -> pd.DataFrame:
        column_fields = list(set(pivot_table.columns) - set(self.label_keys))
        if not self.total_series:
            self.total_series = pivot_table[column_fields].sum(axis=1)

        pivot_table = self._apply_row_sorting(pivot_table)

        if select_fields := self.options.get("select"):
            self._validate_select_fields(select_fields, column_fields)
            pivot_table = pivot_table[self.label_keys + select_fields]
            column_fields = select_fields

        if order_by := self.options.get("order_by"):
            self._validate_order_by_type(order_by)
            pivot_table = self._apply_order_by(pivot_table, order_by, column_fields)

        if limit := self.options.get("limit"):
            if isinstance(limit, float):
                limit = int(limit)
            limited_columns = pivot_table.iloc[
                :, len(self.label_keys) : len(self.label_keys) + limit
            ]
            pivot_table = pivot_table.iloc[:, : len(self.label_keys) + limit]

            pivot_table["Others"] = self.total_series.loc[
                pivot_table.index
            ] - limited_columns.sum(axis=1)

        pivot_table["Sub Total"] = self.total_series.loc[pivot_table.index]
        self.data_keys = [
            col for col in pivot_table.columns if col not in set(self.label_keys)
        ]

        return pivot_table

    @staticmethod
    def validate_labels(labels: dict, df: pd.DataFrame) -> None:
        if not labels:
            raise ERROR_REQUIRED_PARAMETER(key="options.ADD_LABELS.labels")

        for label_key in labels.keys():
            if label_key in df.columns:
                raise ERROR_INVALID_PARAMETER(
                    key="options.ADD_LABELS.labels",
                    reason=f"Duplicated key: {label_key}, columns={list(df.columns)}",
                )

    @staticmethod
    def update_keys(
        key: str,
        value,
        label_keys: list,
        data_keys: list,
    ) -> None:
        if isinstance(value, str):
            label_keys.append(key)
        elif isinstance(value, (int, float)):
            data_keys.append(key)
        else:
            raise ERROR_INVALID_PARAMETER_TYPE(
                key="options.ADD_LABELS.labels", type=type(value)
            )

    def add_labels_to_dataframe(
        self,
        df: pd.DataFrame,
        labels: dict,
        label_keys: list,
        data_keys: list,
    ) -> None:
        for key, value in labels.items():
            df[key] = value
            self.update_keys(key, value, label_keys, data_keys)

    def filter_data(self, df: pd.DataFrame, vars: dict) -> pd.DataFrame:
        condition = self.options.get("condition")
        if not condition:
            return df.copy()

        if self.is_jinja_expression(condition):
            condition, gv_type_map = self.change_global_variables(condition, vars)
            condition = self.remove_jinja_braces(condition)
            condition = self.change_expression_data_type(condition, gv_type_map)
            condition = self.change_space_variable(condition)

        return df.query(condition).copy()

    def apply_cases(self, filtered_df: pd.DataFrame) -> pd.DataFrame:
        name = self.options["name"]
        key = self.options["key"]
        else_value = self.options.get("else")
        cases = self.options.get("cases", [])

        temp_key_column = f"_{key}_backup"
        if temp_key_column not in filtered_df.columns:
            filtered_df[temp_key_column] = filtered_df[key]

        for case in cases:
            self._validate_case(case)
            operator = case["operator"]
            value = case["value"]
            match = case["match"].strip()

            if operator == "eq":
                filtered_df.loc[filtered_df[key] == match, name] = value
            elif operator == "regex":
                filtered_df.loc[
                    filtered_df[key].str.contains(match, na=False), name
                ] = value

        if else_value is not None:
            filtered_df.loc[filtered_df[name].isna(), name] = else_value
        else:
            filtered_df.loc[filtered_df[name].isna(), name] = filtered_df[
                temp_key_column
            ]

        filtered_df.drop(columns=[temp_key_column], inplace=True)

        return filtered_df

    def handle_unfiltered_data(
        self,
        df: pd.DataFrame,
        filtered_df: pd.DataFrame,
        name: str,
        field_type: str,
    ):
        unfiltered_index = df.index.difference(filtered_df.index)

        if field_type == "LABEL":
            df.loc[unfiltered_index, name] = ""
            self.label_keys.append(name)
        elif field_type == "DATA":
            df.loc[unfiltered_index, name] = 0
            self.data_keys.append(name)

    def _apply_row_sorting(
        self,
        pivot_table: pd.DataFrame,
    ) -> pd.DataFrame:
        pivot_table["_total"] = self.total_series
        pivot_table = pivot_table.sort_values(by="_total", ascending=False)
        pivot_table = pivot_table.drop(columns=["_total"])
        return pivot_table

    @staticmethod
    def _validate_order_by_type(order_by: dict) -> None:
        order_by_type = order_by.get("type", "key")
        if order_by_type not in ["key", "value"]:
            raise ERROR_INVALID_PARAMETER(
                key="options.PIVOT.order_by.type",
                reason=f"Invalid order_by type: {order_by_type}",
            )

    def _apply_order_by(
        self,
        pivot_table: pd.DataFrame,
        order_by: dict,
        column_fields: list,
    ) -> pd.DataFrame:
        order_by_type = order_by.get("type", "key")
        desc = order_by.get("desc", False)
        if order_by_type == "key":
            pivot_table = pivot_table[
                self.label_keys + sorted(column_fields, reverse=desc)
            ]
        else:
            column_sums = pivot_table.drop(columns=self.label_keys).sum()
            if desc:
                ascending = False
            else:
                ascending = True
            sorted_columns = column_sums.sort_values(ascending=ascending).index.tolist()
            pivot_table = pivot_table[self.label_keys + sorted_columns]
        return pivot_table

    @staticmethod
    def _validate_case(case):
        required_keys = ["match", "value", "operator"]
        for key in required_keys:
            if key not in case:
                raise ERROR_REQUIRED_PARAMETER(key=f"options.VALUE_MAPPING.cases.{key}")



import logging
from typing import Tuple
from mongoengine import QuerySet

from spaceone.core.manager import BaseManager
from spaceone.dashboard.model.private_widget.database import PrivateWidget
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)

_LOGGER = logging.getLogger(__name__)


class PrivateWidgetManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget_model = PrivateWidget

    def create_private_widget(self, params: dict) -> PrivateWidget:
        def _rollback(vo: PrivateWidget) -> None:
            _LOGGER.info(
                f"[create_private_widget._rollback] "
                f"Delete vo : {vo.name} "
                f"({vo.widget_id})"
            )
            vo.delete()

        widget_vo: PrivateWidget = self.widget_model.create(params)
        self.transaction.add_rollback(_rollback, widget_vo)

        return widget_vo

    def update_private_widget_by_vo(
        self, params: dict, widget_vo: PrivateWidget
    ) -> PrivateWidget:
        def _rollback(old_data: dict) -> None:
            _LOGGER.info(
                f"[update_private_widget_by_vo._rollback] Revert Data : "
                f'{old_data["widget_id"]}'
            )
            widget_vo.update(old_data)

        self.transaction.add_rollback(_rollback, widget_vo.to_dict())
        return widget_vo.update(params)

    @staticmethod
    def delete_private_widget_by_vo(widget_vo: PrivateWidget) -> None:
        # Delete child data tables
        pri_data_table_mgr = PrivateDataTableManager()
        pri_data_table_vos = pri_data_table_mgr.filter_private_data_tables(
            widget_id=widget_vo.widget_id,
            domain_id=widget_vo.domain_id,
        )
        pri_data_table_vos.delete()

        widget_vo.delete()

    def get_private_widget(
        self,
        widget_id: str,
        domain_id: str,
        user_id: str,
    ) -> PrivateWidget:
        conditions = {
            "widget_id": widget_id,
            "domain_id": domain_id,
            "user_id": user_id,
        }

        return self.widget_model.get(**conditions)

    def filter_private_widgets(self, **conditions) -> QuerySet:
        return self.widget_model.filter(**conditions)

    def list_private_widgets(self, query: dict) -> Tuple[QuerySet, int]:
        return self.widget_model.query(**query)

    def stat_private_widgets(self, query: dict) -> dict:
        return self.widget_model.stat(**query)



import logging
import ast
import re
from typing import Union, Literal, Tuple
from jinja2 import Environment, meta
import pandas as pd

from spaceone.core import cache, utils
from spaceone.core.manager import BaseManager
from spaceone.dashboard.error.data_table import (
    ERROR_NO_FIELDS_TO_GLOBAL_VARIABLES,
    ERROR_NOT_GLOBAL_VARIABLE_KEY,
    ERROR_QUERY_OPTION,
    ERROR_QUERY_GROUP_BY_OPTION,
    ERROR_EMPTY_DATA_FIELD,
)

_LOGGER = logging.getLogger(__name__)


class DataTableManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget_id = None
        self.domain_id = None

        self.df: Union[pd.DataFrame, None] = None
        self.data_keys = None
        self.label_keys = None
        self.jinja_variables = None
        self.jinja_variables_contain_space = []
        self.state = None
        self.error_message = None

    def get_data_and_labels_info(self) -> Tuple[dict, dict]:
        raise NotImplementedError()

    def load(
        self,
        granularity: str,
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def load_from_widget(
        self,
        data_table_id: str,
        granularity: str,
        start: str,
        end: str,
        group_by: list = None,
        sort: list = None,
        page: dict = None,
        vars: dict = None,
        column_sum: bool = False,
    ) -> dict:
        query_data = self._prepare_query_data(
            data_table_id, granularity, start, end, group_by, sort
        )

        cache_hash_key = utils.dict_to_hash(query_data)
        response = self._get_cached_response(cache_hash_key)

        if not response:
            self.load(
                granularity,
                start,
                end,
                vars=vars,
            )

            if self.df is not None:
                self._apply_group_by(group_by)
                data_info, labels_info = self.get_data_and_labels_info()

                response = {
                    "results": self.df.copy(deep=True).to_dict(orient="records"),
                }

                if labels_info:
                    response["labels_info"] = labels_info

                if data_info:
                    response["data_info"] = data_info

                if self.data_keys:
                    order = self.label_keys + self.data_keys
                    response["order"] = order

                cache.set(
                    f"dashboard:widget:load:{self.domain_id}:{self.widget_id}:{cache_hash_key}",
                    response,
                    expire=600,
                )

                self.df = None

        if column_sum:
            return self.response_sum_data_from_widget(response)

        return self.response_data_from_widget(response, sort, page)

    def response_data_from_widget(
        self,
        response,
        sort: list = None,
        page: dict = None,
    ) -> dict:
        data = response["results"]

        total_count = len(data)

        if sort:
            data = self.apply_sort(data, sort)

        if page:
            data = self.apply_page(data, page)

        results = {
            "results": data,
            "total_count": total_count,
        }

        if "labels_info" in response:
            results["labels_info"] = response["labels_info"]
        if "data_info" in response:
            results["data_info"] = response["data_info"]
        if "order" in response:
            results["order"] = response["order"]

        return results

    def response_sum_data_from_widget(self, response: dict) -> dict:
        data = response["results"]
        if self.data_keys:
            sum_data = {
                key: sum(float(row.get(key, 0)) for row in data)
                for key in self.data_keys
            }
        else:
            numeric_columns = {
                key
                for row in data
                for key, value in row.items()
                if isinstance(value, (int, float))
            }
            sum_data = {
                key: sum(float(row.get(key, 0)) for row in data)
                for key in numeric_columns
            }

        # results = [{column: sum_value} for column, sum_value in sum_data.items()]
        results = [sum_data]
        return {
            "results": results,
            "total_count": len(results),
        }

    @staticmethod
    def apply_sort(data: list, sort: list) -> list:
        for rule in reversed(sort):
            key = rule["key"]
            reverse = rule.get("desc", False)
            data = sorted(data, key=lambda item: item[key], reverse=reverse)
        return data

    @staticmethod
    def apply_page(data: list, page: dict) -> list:
        if limit := page.get("limit"):
            if limit > 0:
                start = page.get("start", 1)
                if start < 1:
                    start = 1

                start_index = start - 1
                end_index = start_index + limit
                return data[start_index:end_index]

    def response_data(self, sort: list = None, page: dict = None) -> dict:
        total_count = len(self.df)

        if sort:
            self.apply_sort_to_df(sort)

        if page:
            self.apply_page_df(page)

        df = self.df.copy(deep=True)
        data_info, labels_info = self.get_data_and_labels_info()

        self.df = None

        results = {
            "results": df.to_dict(orient="records"),
            "total_count": total_count,
        }

        if labels_info:
            results["labels_info"] = labels_info

        if data_info:
            results["data_info"] = data_info

        if self.data_keys:
            if self.label_keys is None:
                order = self.data_keys
            else:
                order = self.label_keys + self.data_keys

            results["order"] = order

        return results

    def apply_sort_to_df(self, sort: list) -> None:
        if len(self.df) > 0:
            keys = []
            ascendings = []

            for sort_option in sort:
                key = sort_option.get("key")
                ascending = not sort_option.get("desc", False)

                if key:
                    keys.append(key)
                    ascendings.append(ascending)

            try:
                self.df = self.df.sort_values(by=keys, ascending=ascendings)
            except Exception as e:
                _LOGGER.error(f"[_sort] Sort Error: {e}")
                raise ERROR_QUERY_OPTION(key="sort")

    def apply_page_df(self, page: dict) -> None:
        if len(self.df) > 0:
            if limit := page.get("limit"):
                if limit > 0:
                    start = page.get("start", 1)
                    if start < 1:
                        start = 1

                    self.df = self.df.iloc[start - 1 : start + limit - 1]

    def is_jinja_expression(self, expression: str) -> bool:
        if not expression:
            return False

        if isinstance(expression, list):
            expression = str(expression)

        env = Environment()

        jinja_pattern = re.compile(r"\{\{\s*(.*?)\s*\}\}")

        modified_expression = re.sub(
            jinja_pattern,
            lambda m: "{{" + m.group(1).replace(" ", "_") + "}}",
            expression,
        )

        jinja_keys_with_space = {}
        if expression != modified_expression:
            keys_with_space = re.findall(jinja_pattern, expression)
            for key in keys_with_space:
                jinja_keys_with_space[key.replace(" ", "_")] = key

        parsed_content = env.parse(modified_expression)
        variables = meta.find_undeclared_variables(parsed_content)

        if variables:
            if jinja_keys_with_space:
                for key, key_with_space in jinja_keys_with_space.items():
                    self.jinja_variables_contain_space.append(
                        {
                            "origin_key": key_with_space,
                            "modified_key": key,
                        }
                    )

            self.jinja_variables = variables

        return bool(variables)

    def change_global_variables(self, expression: str, vars: dict):
        gv_type_map = {}
        if "global" in self.jinja_variables:
            if not vars:
                raise ERROR_NO_FIELDS_TO_GLOBAL_VARIABLES(vars=vars)

            exclude_keys = set(key for key in self.jinja_variables if key != "global")
            expression = expression.replace("global.", "")

            env = Environment()

            parsed_content = env.parse(expression)
            jinja_variables = meta.find_undeclared_variables(parsed_content)

            global_variables = jinja_variables - exclude_keys
            for global_variable_key in global_variables:
                if global_variable_key not in vars:
                    raise ERROR_NOT_GLOBAL_VARIABLE_KEY(
                        global_variable_key=global_variable_key
                    )

                global_variable_value = vars[global_variable_key]
                gv_type = type(global_variable_value)

                if isinstance(global_variable_value, int) or isinstance(
                    global_variable_value, float
                ):
                    global_variable_value = str(global_variable_value)

                if isinstance(global_variable_value, list):
                    global_variable_value = str(global_variable_value)

                gv_type_map[global_variable_value] = gv_type

                expression = expression.replace(
                    global_variable_key, global_variable_value
                )

        return expression, gv_type_map

    @staticmethod
    def remove_jinja_braces(expression: str) -> Union[str, float, list]:
        jinja_pattern = re.compile(r"\{\{\s*(.*?)\s*\}\}")

        modified_expression = re.sub(
            jinja_pattern,
            lambda m: "{{" + m.group(1).replace(" ", "_") + "}}",
            expression,
        )
        expression = modified_expression

        while "{{" in expression and "}}" in expression:
            if re.match(r"{{\s*(\w+)\s*}}", expression):
                expression = re.sub(r"{{\s*(\w+)\s*}}", r"\1", expression)
            elif re.match(r"{{\s*(\d+(\.\d+)?)\s*}}", expression):
                result = re.sub(r"{{\s*(\d+(\.\d+)?)\s*}}", r"\1", expression)
                try:
                    expression = float(result)
                except ValueError:
                    expression = eval(result)
            else:
                expression = expression.replace("{{", "").replace("}}", "").strip()

                pattern = r"\[\d+\]$"
                if re.search(pattern, expression):
                    dict_expression, _ = re.subn(pattern, "", expression)
                    index_str = expression.replace(dict_expression, "").strip()
                    index_str = index_str.replace("[", "").replace("]", "")
                    index = int(index_str)
                    expression = ast.literal_eval(dict_expression)[index]
                try:
                    expression = ast.literal_eval(expression)
                except Exception as e:
                    _LOGGER.error(f"Error: {e}")
                    expression = expression
        return expression

    @staticmethod
    def change_expression_data_type(expression: str, gv_type_map: dict) -> str:
        for gv_value, data_type in gv_type_map.items():
            if isinstance(data_type(gv_value), str):
                expression = expression.replace(gv_value, f'"{gv_value}"')

        return expression

    def change_space_variable(self, expression: str) -> str:
        if self.jinja_variables_contain_space:
            for space_variable in self.jinja_variables_contain_space:
                expression = expression.replace(
                    space_variable["modified_key"], f"`{space_variable['origin_key']}`"
                )

        return expression

    def _prepare_query_data(
        self,
        data_table_id: str,
        granularity: str,
        start: str,
        end: str,
        group_by: list,
        sort: list,
    ) -> dict:
        user_id = self.transaction.get_meta(
            "authorization.user_id"
        ) or self.transaction.get_meta("authorization.app_id")
        role_type = self.transaction.get_meta("authorization.role_type")

        query_data = {
            "granularity": granularity,
            "start": start,
            "end": end,
            "group_by": group_by,
            "sort": sort,
            "data_table_id": data_table_id,
            "widget_id": self.widget_id,
            "domain_id": self.domain_id,
        }

        if role_type == "WORKSPACE_MEMBER":
            query_data["user_id"] = user_id

        return query_data

    def _get_cached_response(self, cache_hash_key) -> dict:
        cache_data = cache.get(
            f"dashboard:widget:load:{self.domain_id}:{self.widget_id}:{cache_hash_key}"
        )
        return cache_data if cache_data else None

    def _apply_group_by(self, group_by: list):
        if group_by and not self.df.empty:
            for key in group_by:
                if key not in self.df.columns:
                    raise ERROR_QUERY_GROUP_BY_OPTION(
                        key="group_by", fields=list(self.df.columns)
                    )

            agg_funcs = {
                column: "sum"
                for column in self.df.columns
                if pd.api.types.is_numeric_dtype(self.df[column])
            }

            if not agg_funcs:
                raise ERROR_EMPTY_DATA_FIELD(fields=list(self.df.columns))

            self.df = self.df.groupby(group_by).agg(agg_funcs).reset_index()



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.dashboard.manager.private_widget_manager import PrivateWidgetManager
from spaceone.dashboard.manager.private_dashboard_manager import PrivateDashboardManager
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.data_table_manager.data_transformation_manager import (
    DataTransformationManager,
)
from spaceone.dashboard.model.private_widget.request import *
from spaceone.dashboard.model.private_widget.response import *
from spaceone.dashboard.error.dashboard import (
    ERROR_NOT_SUPPORTED_VERSION,
    ERROR_INVALID_PARAMETER,
    ERROR_REQUIRED_PARAMETER,
)
from spaceone.dashboard.service.private_data_table_service import (
    PrivateDataTableService,
)

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PrivateWidgetService(BaseService):
    resource = "PrivateWidget"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pri_widget_mgr = PrivateWidgetManager()
        self.pri_data_table_svc = PrivateDataTableService()
        self.data_table_id_map = {}

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def create(
        self, params: PrivateWidgetCreateRequest
    ) -> Union[PrivateWidgetResponse, dict]:
        """Create private widget

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'name': 'str',
                'state': 'str',
                'description': 'str',
                'widget_type': 'str',
                'size': 'str',
                'options': 'dict',
                'data_table_id': 'int',
                'data_tables': 'list',
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str',             # injected from auth (required)
            }

        Returns:
            PrivateWidgetResponse:
        """

        params_dict = params.dict()
        if "data_tables" in params_dict:
            params_dict["is_bulk"] = True

        pri_widget_info = self.create_widget(params_dict)
        return PrivateWidgetResponse(**pri_widget_info)

    @check_required(["dashboard_id", "domain_id", "user_id"])
    def create_widget(self, params_dict: dict) -> dict:
        dashboard_id = params_dict["dashboard_id"]
        domain_id = params_dict["domain_id"]
        user_id = params_dict["user_id"]
        is_bulk = params_dict.get("is_bulk", False)
        data_table_idx = params_dict.get("data_table_id") or 0
        data_tables = params_dict.get("data_tables")
        self.data_table_id_map = {}

        pri_dashboard_mgr = PrivateDashboardManager()
        pri_dashboard_vo = pri_dashboard_mgr.get_private_dashboard(
            dashboard_id, domain_id, user_id
        )

        if pri_dashboard_vo.version == "1.0":
            raise ERROR_NOT_SUPPORTED_VERSION(version=pri_dashboard_vo.version)

        if is_bulk and data_table_idx is not None:
            del params_dict["data_table_id"]

        pri_widget_vo = self.pri_widget_mgr.create_private_widget(params_dict)

        if is_bulk and data_tables:
            self._create_data_tables(
                data_tables,
                pri_widget_vo.widget_id,
                domain_id,
                user_id,
            )

            widget_data_table_id = self.data_table_id_map[data_table_idx]
            self.pri_widget_mgr.update_private_widget_by_vo(
                {"data_table_id": widget_data_table_id}, pri_widget_vo
            )

        return pri_widget_vo.to_dict()

    def _create_data_tables(
        self,
        data_tables: list,
        widget_id: str,
        domain_id: str,
        user_id: str,
    ):
        retry_data_tables = {}
        idx = 0
        for data_table in data_tables:
            data_table["widget_id"] = widget_id
            data_table["domain_id"] = domain_id
            data_table["user_id"] = user_id
            data_table["is_bulk"] = True

            if data_type := data_table.get("data_type"):
                if data_type == "ADDED":
                    pub_data_table_info = self.pri_data_table_svc.add_data_table(
                        data_table
                    )
                elif data_type == "TRANSFORMED":
                    operator = data_table.get("operator")
                    if operator is None:
                        raise ERROR_REQUIRED_PARAMETER(
                            key="layouts.widgets.data_tables.operator"
                        )

                    options = data_table.get("options", {})
                    operator_options = options.get(operator, {})

                    if "data_table_id" in operator_options:
                        data_table_idx = operator_options["data_table_id"]
                        if data_table_idx in self.data_table_id_map:
                            data_table["options"][operator]["data_table_id"] = (
                                self.data_table_id_map[data_table_idx]
                            )
                        else:
                            retry_data_tables[idx] = data_table
                            idx += 1
                            continue

                    if "data_tables" in operator_options:
                        data_tables = operator_options["data_tables"]
                        changed_data_tables = []
                        is_success = True
                        for data_table_idx in data_tables:
                            if data_table_idx in self.data_table_id_map:
                                changed_data_tables.append(
                                    self.data_table_id_map[data_table_idx]
                                )
                            else:
                                is_success = False
                                break
                        if is_success:
                            data_table["options"][operator][
                                "data_tables"
                            ] = changed_data_tables
                        else:
                            retry_data_tables[idx] = data_table
                            idx += 1
                            continue

                    pub_data_table_info = self.pri_data_table_svc.transform_data_table(
                        data_table
                    )
                else:
                    raise ERROR_INVALID_PARAMETER(
                        key="layouts.widgets.data_tables.data_type",
                        reason=f"Data type is not supported. (data_type = {data_type})",
                    )
            else:
                raise ERROR_REQUIRED_PARAMETER(
                    key="layouts.widgets.data_tables.data_type"
                )

            created_data_table_id = pub_data_table_info["data_table_id"]

            _LOGGER.debug(f"[create_widget] create data table: {created_data_table_id}")
            self.data_table_id_map[idx] = created_data_table_id
            idx += 1

        if retry_data_tables:
            self._retry_create_data_tables(retry_data_tables)

    def _retry_create_data_tables(self, data_tables: dict):
        retry_data_tables = {}
        for idx, data_table in data_tables.items():
            operator = data_table.get("operator")
            options = data_table.get("options", {})
            operator_options = options.get(operator, {})

            if "data_table_id" in operator_options:
                data_table_idx = operator_options["data_table_id"]
                if data_table_idx in self.data_table_id_map:
                    data_table["options"][operator]["data_table_id"] = (
                        self.data_table_id_map[data_table_idx]
                    )
                else:
                    retry_data_tables[idx] = data_table
                    idx += 1
                    continue

            if "data_tables" in operator_options:
                data_tables = operator_options["data_tables"]
                changed_data_tables = []
                is_success = True
                for data_table_idx in data_tables:
                    if data_table_idx in self.data_table_id_map:
                        changed_data_tables.append(
                            self.data_table_id_map[data_table_idx]
                        )
                    else:
                        is_success = False
                        break
                if is_success:
                    data_table["options"][operator]["data_tables"] = changed_data_tables
                else:
                    retry_data_tables[idx] = data_table
                    idx += 1
                    continue

            pub_data_table_info = self.pri_data_table_svc.transform_data_table(
                data_table
            )
            created_data_table_id = pub_data_table_info["data_table_id"]

            _LOGGER.debug(f"[create_widget] create data table: {created_data_table_id}")
            self.data_table_id_map[idx] = created_data_table_id

        if retry_data_tables:
            self._retry_create_data_tables(retry_data_tables)

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def update(
        self, params: PrivateWidgetUpdateRequest
    ) -> Union[PrivateWidgetResponse, dict]:
        """Update private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'state': 'str',
                'description': 'str',
                'widget_type': 'str',
                'size': 'str',
                'options': 'dict',
                'data_table_id': 'str',
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateWidgetResponse:
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        if params.data_table_id is not None:
            pri_data_table_mgr = PrivateDataTableManager()
            pri_data_table_vo = pri_data_table_mgr.get_private_data_table(
                params.data_table_id,
                params.domain_id,
                params.user_id,
            )

            if pri_data_table_vo.widget_id != params.widget_id:
                raise ERROR_INVALID_PARAMETER(
                    key="data_table_id",
                    reason="Data table is not belong to this widget.",
                )

        pri_widget_vo = self.pri_widget_mgr.update_private_widget_by_vo(
            params.dict(exclude_unset=True), pri_widget_vo
        )

        return PrivateWidgetResponse(**pri_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def delete(self, params: PrivateWidgetDeleteRequest) -> None:
        """Delete private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        self.pri_widget_mgr.delete_private_widget_by_vo(pri_widget_vo)

    @transaction(
        permission="dashboard:PrivateWidget.read",
        role_types=["USER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "project_id", "*")
    @convert_model
    def load(self, params: PrivateWidgetLoadRequest) -> dict:
        """Load private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'granularity': 'str',           # required
                'start': 'str',                 # required
                'end': 'str',                   # required
                'group_by': 'list',
                'sort': 'list',
                'page': 'dict',
                'vars': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        if pri_widget_vo.data_table_id is None:
            raise ERROR_INVALID_PARAMETER(
                key="widget_id", reason="Data table is not set."
            )

        pri_data_table_mgr = PrivateDataTableManager()
        pri_data_table_vo = pri_data_table_mgr.get_private_data_table(
            pri_widget_vo.data_table_id,
            params.domain_id,
            params.user_id,
        )

        if pri_data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                "PRIVATE",
                pri_data_table_vo.source_type,
                pri_data_table_vo.options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            return ds_mgr.load_from_widget(
                pri_data_table_vo.data_table_id,
                params.granularity,
                params.start,
                params.end,
                params.group_by,
                params.sort,
                params.page,
                params.vars,
            )

        else:
            operator = pri_data_table_vo.operator
            options = pri_data_table_vo.options.get(operator, {})

            dt_mgr = DataTransformationManager(
                "PRIVATE",
                pri_data_table_vo.operator,
                options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            return dt_mgr.load_from_widget(
                pri_data_table_vo.data_table_id,
                params.granularity,
                params.start,
                params.end,
                params.group_by,
                params.sort,
                params.page,
                params.vars,
            )

    @transaction(
        permission="dashboard:PrivateWidget.read",
        role_types=["USER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "project_id", "*")
    @convert_model
    def load_sum(self, params: PrivateWidgetLoadSumRequest) -> dict:
        """Load private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'granularity': 'str',           # required
                'start': 'str',                 # required
                'end': 'str',                   # required
                'vars': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        if pri_widget_vo.data_table_id is None:
            raise ERROR_INVALID_PARAMETER(
                key="widget_id", reason="Data table is not set."
            )

        pri_data_table_mgr = PrivateDataTableManager()
        pri_data_table_vo = pri_data_table_mgr.get_private_data_table(
            pri_widget_vo.data_table_id,
            params.domain_id,
            params.user_id,
        )

        if pri_data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                "PRIVATE",
                pri_data_table_vo.source_type,
                pri_data_table_vo.options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            return ds_mgr.load_from_widget(
                pri_data_table_vo.data_table_id,
                params.granularity,
                params.start,
                params.end,
                vars=params.vars,
                column_sum=True,
            )

        else:
            operator = pri_data_table_vo.operator
            options = pri_data_table_vo.options.get(operator, {})

            dt_mgr = DataTransformationManager(
                "PRIVATE",
                pri_data_table_vo.operator,
                options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            return dt_mgr.load_from_widget(
                pri_data_table_vo.data_table_id,
                params.granularity,
                params.start,
                params.end,
                vars=params.vars,
                column_sum=True,
            )

    @transaction(
        permission="dashboard:PrivateWidget.read",
        role_types=["USER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "project_id", "*")
    @convert_model
    def get(
        self, params: PrivateWidgetGetRequest
    ) -> Union[PrivateWidgetResponse, dict]:
        """Get private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateWidgetResponse:
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        return PrivateWidgetResponse(**pri_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateWidget.read",
        role_types=["USER"],
    )
    @append_query_filter(
        [
            "dashboard_id",
            "widget_id",
            "name",
            "domain_id",
            "user_id",
        ]
    )
    @append_keyword_filter(["widget_id", "name"])
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "project_id", "*")
    @convert_model
    def list(
        self, params: PrivateWidgetSearchQueryRequest
    ) -> Union[PrivateWidgetsResponse, dict]:
        """List private widgets

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',                          # required
                'widget_id': 'str',
                'name': 'str',
                'user_id': 'str',                               # injected from auth (required)
                'domain_id': 'str',                             # injected from auth (required)
            }

        Returns:
            PrivateWidgetsResponse:
        """

        query = params.query or {}
        pri_widget_vos, total_count = self.pri_widget_mgr.list_private_widgets(query)
        pri_widgets_info = [pri_widget_vo.to_dict() for pri_widget_vo in pri_widget_vos]
        return PrivateWidgetsResponse(results=pri_widgets_info, total_count=total_count)



import logging
import copy
from typing import Tuple
from datetime import datetime
from dateutil.relativedelta import relativedelta
import pandas as pd

from spaceone.dashboard.manager.data_table_manager import DataTableManager
from spaceone.dashboard.manager.cost_analysis_manager import CostAnalysisManager
from spaceone.dashboard.manager.inventory_manager import InventoryManager
from spaceone.dashboard.error.data_table import *

_LOGGER = logging.getLogger(__name__)


class DataSourceManager(DataTableManager):
    def __init__(
        self,
        data_table_type: str,
        source_type: str,
        options: dict,
        widget_id: str,
        domain_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if source_type not in ["COST", "ASSET"]:
            raise ERROR_NOT_SUPPORTED_SOURCE_TYPE(source_type=source_type)

        self.data_table_type = data_table_type
        self.cost_analysis_mgr = CostAnalysisManager()
        self.inventory_mgr = InventoryManager()
        self.source_type = source_type
        self.options = options
        self.widget_id = widget_id
        self.domain_id = domain_id

        self.data_name = options.get("data_name")
        if self.data_name is None:
            raise ERROR_REQUIRED_PARAMETER(key="options.data_name")

        self.data_unit = options.get("data_unit")
        self.timediff = options.get("timediff")
        if self.timediff:
            self.timediff_data_name = self.timediff["data_name"]

        self.group_by = options.get("group_by")
        self.filter = options.get("filter")
        self.filter_or = options.get("filter_or")

    def get_data_and_labels_info(self) -> Tuple[dict, dict]:
        data_info = {self.data_name: {}}

        if self.data_unit:
            data_info[self.data_name]["unit"] = self.data_unit

        if self.timediff:
            data_info[self.timediff["data_name"]] = {}

        labels_info = {}
        if self.group_by:
            for group_option in copy.deepcopy(self.group_by):
                if isinstance(group_option, dict):
                    group_name = group_option.get("name")
                    group_key = group_option.get("key")
                    if "." in group_key:
                        group_key = group_key.split(".")[-1]

                    name = group_name or group_key
                    if name is None:
                        raise ERROR_REQUIRED_PARAMETER(key="options.group_by.key")

                    if group_name:
                        del group_option["name"]

                    if group_key:
                        del group_option["key"]

                    labels_info[group_name] = group_option
                else:
                    labels_info[group_option] = {}

        labels_info["Date"] = {}

        return data_info, labels_info

    def load(
        self,
        granularity: str = "MONTHLY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> pd.DataFrame:
        try:
            start, end = self._get_time_from_granularity(granularity, start, end)

            if self.source_type == "COST":
                self._analyze_cost(granularity, start, end, vars)
            elif self.source_type == "ASSET":
                self._analyze_asset(granularity, start, end, vars)

            if self.timediff:
                self.df = self._apply_timediff(granularity, start, end, vars)

            if self.group_by:
                self._add_none_value_group_by_columns()

            self.state = "AVAILABLE"
            self.error_message = None

        except Exception as e:
            self.state = "UNAVAILABLE"
            self.error_message = e.message if hasattr(e, "message") else str(e)
            _LOGGER.error(f"[load] add {self.source_type} source error: {e}")

        return self.df

    def _analyze_asset(
        self,
        granularity: str,
        start: str,
        end: str,
        vars: dict = None,
    ) -> None:
        asset_info = self.options.get("ASSET", {})
        metric_id = asset_info.get("metric_id")
        data_key = "value"

        if metric_id is None:
            raise ERROR_REQUIRED_PARAMETER(parameter="options.ASSET.metric_id")

        query = self._make_query(
            data_key,
            granularity,
            start,
            end,
            vars=vars,
        )

        params = {"metric_id": metric_id, "query": query}

        response = self.inventory_mgr.analyze_metric_data(params)
        results = response.get("results", [])

        results = self._change_datetime_format(results)

        self.df = pd.DataFrame(results)

    def _analyze_cost(
        self,
        granularity: str,
        start: str,
        end: str,
        vars: dict = None,
    ) -> None:
        cost_info = self.options.get("COST", {})
        data_source_id = cost_info.get("data_source_id")
        data_key = cost_info.get("data_key")

        if data_source_id is None:
            raise ERROR_REQUIRED_PARAMETER(parameter="options.COST.data_source_id")

        if data_key is None:
            raise ERROR_REQUIRED_PARAMETER(parameter="options.COST.data_key")

        query = self._make_query(
            data_key,
            granularity,
            start,
            end,
            vars=vars,
        )

        params = {"data_source_id": data_source_id, "query": query}

        response = self.cost_analysis_mgr.analyze_cost(params)
        results = response.get("results", [])

        results = self._change_datetime_format(results)

        self.df = pd.DataFrame(results)

    def _apply_timediff(
        self, granularity: str, start: str, end: str, vars: dict
    ) -> pd.DataFrame:
        origin_df = self.df.copy()

        start, end = self._change_query_time(granularity, start, end)

        if self.source_type == "COST":
            self._analyze_cost(granularity, start, end, vars)
        elif self.source_type == "ASSET":
            self._analyze_asset(granularity, start, end, vars)

        if self.df.empty:
            _LOGGER.debug(
                "self.df is empty, creating a DataFrame with the same columns as origin_df."
            )
            self.df = pd.DataFrame(columns=origin_df.columns)

        self.df["Date"] = self.df["Date"].apply(
            lambda x: self._change_date_by_timediff(x)
        )
        self.df.rename(columns={self.data_name: self.timediff_data_name}, inplace=True)

        origin_label_keys = [
            column for column in origin_df.columns if column != self.data_name
        ]
        diff_label_keys = [
            column for column in self.df.columns if column != self.timediff_data_name
        ]
        join_keys = list(set(origin_label_keys) & set(diff_label_keys))

        merged_df = pd.merge(origin_df, self.df, on=join_keys, how="outer")

        fill_na = {key: 0 for key in [self.data_name, self.timediff_data_name]}
        fill_na.update({key: "" for key in join_keys})
        merged_df = merged_df.fillna(value=fill_na)

        return merged_df

    @staticmethod
    def _change_datetime_format(results: list) -> list:
        changed_results = []
        for result in results:
            if date := result.get("date"):
                result["Date"] = date
                del result["date"]
            changed_results.append(result)
        return changed_results

    def _change_date_by_timediff(self, date: str) -> str:
        dt = self._get_datetime_from_str(date)
        years = int(self.timediff.get("years", 0))
        months = int(self.timediff.get("months", 0))

        if years:
            dt = dt - relativedelta(years=years)
        elif months:
            dt = dt - relativedelta(months=months)

        return self._change_str_from_datetime(dt, len(date))

    def _change_query_time(
        self, granularity: str, start: str, end: str
    ) -> Tuple[str, str]:
        start_len = len(start)
        end_len = len(end)
        start_time = self._get_datetime_from_str(start)
        end_time = self._get_datetime_from_str(end, is_end=True)

        years = self.timediff.get("years", 0)
        months = self.timediff.get("months", 0)

        if years:
            start_time = start_time + relativedelta(years=years)
            end_time = end_time + relativedelta(years=years)
        elif months:
            start_time = start_time + relativedelta(months=months)
            end_time = end_time + relativedelta(months=months)

        if granularity == "YEARLY":
            if start_len == 4:
                if start_time + relativedelta(years=3) <= end_time:
                    end_time = start_time + relativedelta(years=2)
        elif granularity == "MONTHLY":
            if start_len == 4:
                start_time = end_time
            elif start_len == 7:
                if start_time + relativedelta(months=12) <= end_time:
                    end_time = start_time + relativedelta(months=11)
        else:
            if start_len <= 7:
                start_time = end_time
            elif start_len == 10:
                if start_time + relativedelta(months=1) <= end_time:
                    end_time = (
                        start_time + relativedelta(months=1) - relativedelta(days=1)
                    )

        return (
            self._change_str_from_datetime(start_time, start_len),
            self._change_str_from_datetime(end_time, end_len),
        )

    @staticmethod
    def _change_str_from_datetime(dt: datetime, date_str_len: int) -> str:
        if date_str_len == 4:
            return dt.strftime("%Y")
        elif date_str_len == 7:
            return dt.strftime("%Y-%m")
        else:
            return dt.strftime("%Y-%m-%d")

    @staticmethod
    def _get_datetime_from_str(datetime_str: str, is_end: bool = False) -> datetime:
        if len(datetime_str) == 4:
            dt = datetime.strptime(datetime_str, "%Y")
            if is_end:
                dt = dt + relativedelta(years=1) - relativedelta(days=1)

        elif len(datetime_str) == 7:
            dt = datetime.strptime(datetime_str, "%Y-%m")
            if is_end:
                dt = dt + relativedelta(months=1) - relativedelta(days=1)
        else:
            dt = datetime.strptime(datetime_str, "%Y-%m-%d")

        return dt

    @staticmethod
    def _get_time_from_granularity(
        granularity: str,
        start: str = None,
        end: str = None,
    ) -> Tuple[str, str]:
        if start and end:
            return start, end
        else:
            now = datetime.utcnow()

            if granularity == "YEARLY":
                end_time = now.replace(month=1, day=1, hour=0, minute=0, second=0)
                start_time = end_time - relativedelta(years=2)
                return start_time.strftime("%Y"), end_time.strftime("%Y")

            elif granularity == "MONTHLY":
                end_time = now.replace(day=1, hour=0, minute=0, second=0)
                start_time = end_time - relativedelta(months=5)
                return start_time.strftime("%Y-%m"), end_time.strftime("%Y-%m")

            else:
                end_time = now.replace(hour=0, minute=0, second=0)
                start_time = end_time - relativedelta(days=29)
                return start_time.strftime("%Y-%m-%d"), end_time.strftime("%Y-%m-%d")

    def _make_query(
        self,
        data_key: str,
        granularity: str,
        start: str,
        end: str,
        vars: dict = None,
    ):
        if self.filter:
            for filter_info in self.filter:
                query_value = filter_info.get("v") or filter_info.get("value")
                if self.is_jinja_expression(query_value):
                    query_value, gv_type_map = self.change_global_variables(
                        query_value, vars
                    )
                    query_value = self.remove_jinja_braces(query_value)
                    if (
                        isinstance(query_value, str)
                        or isinstance(query_value, int)
                        or isinstance(query_value, float)
                    ):
                        filter_info["v"] = [query_value]
                    elif isinstance(query_value, list):
                        filter_info["v"] = query_value

        return {
            "granularity": granularity,
            "start": start,
            "end": end,
            "group_by": self.group_by,
            "filter": self.filter,
            "filter_or": self.filter_or,
            "fields": {self.data_name: {"key": data_key, "operator": "sum"}},
        }

    def _add_none_value_group_by_columns(self) -> None:
        group_by_columns = [group_option.get("name") for group_option in self.group_by]
        none_group_by_columns = list(set(group_by_columns) - set(self.df.columns))

        for column in none_group_by_columns:
            self.df[column] = None



from spaceone.core.pygrpc.server import GRPCServer
from spaceone.dashboard.interface.grpc.public_folder import PublicFolder
from spaceone.dashboard.interface.grpc.public_dashboard import PublicDashboard
from spaceone.dashboard.interface.grpc.public_widget import PublicWidget
from spaceone.dashboard.interface.grpc.public_data_table import PublicDataTable
from spaceone.dashboard.interface.grpc.private_folder import PrivateFolder
from spaceone.dashboard.interface.grpc.private_dashboard import PrivateDashboard
from spaceone.dashboard.interface.grpc.private_widget import PrivateWidget
from spaceone.dashboard.interface.grpc.private_data_table import PrivateDataTable

__all__ = ["app"]

app = GRPCServer()
app.add_service(PublicFolder)
app.add_service(PublicDashboard)
app.add_service(PublicWidget)
app.add_service(PublicDataTable)
app.add_service(PrivateFolder)
app.add_service(PrivateDashboard)
app.add_service(PrivateWidget)
app.add_service(PrivateDataTable)



import logging
import copy
from typing import Union

from spaceone.core import cache
from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.error.data_table import ERROR_UNAVAILABLE_DATA_TABLE
from spaceone.dashboard.manager.public_data_table_manager import PublicDataTableManager
from spaceone.dashboard.manager.public_widget_manager import PublicWidgetManager
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.data_table_manager.data_transformation_manager import (
    DataTransformationManager,
)
from spaceone.dashboard.manager.cost_analysis_manager import CostAnalysisManager
from spaceone.dashboard.model.public_data_table.request import *
from spaceone.dashboard.model.public_data_table.response import *
from spaceone.dashboard.model.public_data_table.database import PublicDataTable

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PublicDataTableService(BaseService):
    resource = "PublicDataTable"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pub_data_table_mgr = PublicDataTableManager()

    @transaction(
        permission="dashboard:PublicDataTable.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def add(
        self, params: PublicDataTableAddRequest
    ) -> Union[PublicDataTableResponse, dict]:
        """Add public data table

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'source_type': 'str',           # required
                'options': 'dict',              # required
                'vars': 'dict',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str',             # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicDataTableResponse:
        """

        pub_data_table_info = self.add_data_table(params.dict())
        return PublicDataTableResponse(**pub_data_table_info)

    @check_required(["widget_id", "source_type", "options", "domain_id"])
    def add_data_table(self, params_dict: dict) -> dict:
        source_type = params_dict.get("source_type")
        options = params_dict.get("options")
        vars = params_dict.get("vars")
        widget_id = params_dict.get("widget_id")
        domain_id = params_dict.get("domain_id")
        workspace_id = params_dict.get("workspace_id")
        user_projects = params_dict.get("user_projects")

        pub_widget_mgr = PublicWidgetManager()
        pub_widget_vo = pub_widget_mgr.get_public_widget(
            widget_id,
            domain_id,
            workspace_id,
            user_projects,
        )

        raw_filter = copy.deepcopy(options.get("filter"))

        if source_type == "COST":
            if plugin_id := options.get("COST", {}).get("plugin_id"):
                data_source_id = self._get_data_source_id_from_plugin_id(plugin_id)
                options["COST"]["data_source_id"] = data_source_id
                params_dict["options"] = options
                del params_dict["options"]["COST"]["plugin_id"]

        ds_mgr = DataSourceManager(
            "PUBLIC",
            source_type,
            options,
            widget_id,
            domain_id,
        )

        # Load data source to verify options
        ds_mgr.load(vars=vars)

        # Get data and labels info from options
        data_info, labels_info = ds_mgr.get_data_and_labels_info()

        params_dict["data_type"] = "ADDED"
        params_dict["data_info"] = data_info
        params_dict["labels_info"] = labels_info
        params_dict["dashboard_id"] = pub_widget_vo.dashboard_id
        params_dict["resource_group"] = pub_widget_vo.resource_group
        params_dict["workspace_id"] = pub_widget_vo.workspace_id
        params_dict["project_id"] = pub_widget_vo.project_id
        params_dict["state"] = ds_mgr.state
        params_dict["error_message"] = ds_mgr.error_message

        if raw_filter:
            params_dict["options"]["filter"] = raw_filter

        pub_data_table_vo = self.pub_data_table_mgr.create_public_data_table(
            params_dict
        )

        return pub_data_table_vo.to_dict()

    @staticmethod
    def _get_data_source_id_from_plugin_id(plugin_id: str) -> str:
        cost_mgr = CostAnalysisManager()
        params = {
            "query": {
                "filter": [{"k": "plugin_info.plugin_id", "v": plugin_id, "o": "eq"}],
            }
        }
        data_sources_info = cost_mgr.list_data_sources(params)
        if data_sources_info.get("total_count", 0) == 0:
            raise ERROR_INVALID_PARAMETER(
                key="options.COST.plugin_id",
                reason=f"Invalid plugin_id: {plugin_id}",
            )
        data_source_info = data_sources_info.get("results")[0]
        return data_source_info.get("data_source_id")

    @transaction(
        permission="dashboard:PublicDataTable.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def transform(
        self, params: PublicDataTableTransformRequest
    ) -> Union[PublicDataTableResponse, dict]:
        """Transform public data table

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'operator': 'str',              # required
                'options': 'dict',              # required
                'vars': 'dict',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str',             # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicDataTableResponse:
        """

        pub_data_table_info = self.transform_data_table(params.dict())
        return PublicDataTableResponse(**pub_data_table_info)

    @check_required(["widget_id", "operator", "options", "domain_id"])
    def transform_data_table(self, params_dict: dict) -> dict:
        operator = params_dict.get("operator")
        options = params_dict.get("options")
        operator_options = options.get(operator, {})
        vars = params_dict.get("vars")
        widget_id = params_dict.get("widget_id")
        domain_id = params_dict.get("domain_id")
        workspace_id = params_dict.get("workspace_id")
        user_projects = params_dict.get("user_projects")

        pub_widget_mgr = PublicWidgetManager()
        pub_widget_vo = pub_widget_mgr.get_public_widget(
            widget_id,
            domain_id,
            workspace_id,
            user_projects,
        )

        dt_mgr = DataTransformationManager(
            "PUBLIC",
            operator,
            operator_options,
            widget_id,
            domain_id,
        )

        # Load data table to verify options
        dt_mgr.load(vars=vars)

        # Get data and labels info from options
        data_info, labels_info = dt_mgr.get_data_and_labels_info()

        params_dict["data_type"] = "TRANSFORMED"
        params_dict["data_info"] = data_info
        params_dict["labels_info"] = labels_info
        params_dict["dashboard_id"] = pub_widget_vo.dashboard_id
        params_dict["resource_group"] = pub_widget_vo.resource_group
        params_dict["workspace_id"] = pub_widget_vo.workspace_id
        params_dict["project_id"] = pub_widget_vo.project_id
        params_dict["state"] = dt_mgr.state
        params_dict["error_message"] = dt_mgr.error_message

        pub_data_table_vo = self.pub_data_table_mgr.create_public_data_table(
            params_dict
        )

        return pub_data_table_vo.to_dict()

    @transaction(
        permission="dashboard:PublicDataTable.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def update(
        self, params: PublicDataTableUpdateRequest
    ) -> Union[PublicDataTableResponse, dict]:
        """Update public data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'name': 'str',
                'options': 'dict',
                'vars': 'dict',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicDataTableResponse:
        """

        pub_data_table_vo: PublicDataTable = (
            self.pub_data_table_mgr.get_public_data_table(
                params.data_table_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )
        )

        params_dict = params.dict(exclude_unset=True)
        vars = params_dict.get("vars")

        if options := params_dict.get("options"):
            raw_filter = copy.deepcopy(options.get("filter"))

            if pub_data_table_vo.data_type == "ADDED":
                ds_mgr = DataSourceManager(
                    "PUBLIC",
                    pub_data_table_vo.source_type,
                    options,
                    pub_data_table_vo.widget_id,
                    pub_data_table_vo.domain_id,
                )

                # Load data source to verify options
                ds_mgr.load(vars=vars)

                # Get ds_mgr state and error_message
                params_dict["state"] = ds_mgr.state
                params_dict["error_message"] = ds_mgr.error_message

                # Get data and labels info from options
                data_info, labels_info = ds_mgr.get_data_and_labels_info()
                params_dict["data_info"] = data_info
                params_dict["labels_info"] = labels_info

                # change timediff format
                if timediff := options.get("timediff"):
                    if years := timediff.get("years"):
                        options["timediff"] = {"years": years}
                    elif months := timediff.get("months"):
                        options["timediff"] = {"months": months}

                    if data_name := timediff.get("data_name"):
                        options["timediff"].update({"data_name": data_name})

                    params_dict["options"] = options
            else:
                operator = pub_data_table_vo.operator
                operator_options = options.get(operator, {})
                dt_mgr = DataTransformationManager(
                    "PUBLIC",
                    operator,
                    operator_options,
                    pub_data_table_vo.widget_id,
                    pub_data_table_vo.domain_id,
                )

                # Load data table to verify options
                dt_mgr.load(vars=vars)

                # Get data and labels info from options
                data_info, labels_info = dt_mgr.get_data_and_labels_info()

                params_dict = params.dict()
                params_dict["data_type"] = "TRANSFORMED"
                params_dict["data_info"] = data_info
                params_dict["labels_info"] = labels_info
                params_dict["options"] = {
                    operator: operator_options,
                }

                # Get dt_mgr state and error_message
                params_dict["state"] = dt_mgr.state
                params_dict["error_message"] = dt_mgr.error_message

            if raw_filter:
                params_dict["options"]["filter"] = raw_filter

        cache_key = (
            f"dashboard:widget:load:{params.domain_id}:{pub_data_table_vo.widget_id}:*"
        )
        cache.delete_pattern(cache_key)

        pub_data_table_vo = self.pub_data_table_mgr.update_public_data_table_by_vo(
            params_dict, pub_data_table_vo
        )

        return PublicDataTableResponse(**pub_data_table_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDataTable.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def delete(self, params: PublicDataTableDeleteRequest) -> None:
        """Delete public data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_data_table_vo: PublicDataTable = (
            self.pub_data_table_mgr.get_public_data_table(
                params.data_table_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )
        )

        cache_key = (
            f"dashboard:Widget:load:{params.domain_id}:{pub_data_table_vo.widget_id}:*"
        )
        cache.delete_pattern(cache_key)

        self.pub_data_table_mgr.delete_public_data_table_by_vo(pub_data_table_vo)

    @transaction(
        permission="dashboard:PublicDataTable.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "user_projects", "*")
    @convert_model
    def load(self, params: PublicDataTableLoadRequest) -> dict:
        """Load public data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'granularity': 'str',           # required
                'start': 'str',
                'end': 'str',
                'sort': 'list',
                'page': 'dict',
                'vars': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_data_table_vo: PublicDataTable = (
            self.pub_data_table_mgr.get_public_data_table(
                params.data_table_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )
        )

        if pub_data_table_vo.state == "UNAVAILABLE":
            raise ERROR_UNAVAILABLE_DATA_TABLE(
                data_table_id=pub_data_table_vo.data_table_id
            )

        if pub_data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                "PUBLIC",
                pub_data_table_vo.source_type,
                pub_data_table_vo.options,
                pub_data_table_vo.widget_id,
                pub_data_table_vo.domain_id,
            )
            ds_mgr.load(
                params.granularity,
                params.start,
                params.end,
                params.vars,
            )

            if ds_mgr.state == "UNAVAILABLE":
                raise ERROR_UNAVAILABLE_DATA_TABLE(
                    data_table_id=pub_data_table_vo.data_table_id
                )

            return ds_mgr.response_data(params.sort, params.page)

        else:
            operator = pub_data_table_vo.operator
            options = pub_data_table_vo.options.get(operator, {})
            widget_id = pub_data_table_vo.widget_id
            domain_id = pub_data_table_vo.domain_id

            dt_mgr = DataTransformationManager(
                "PUBLIC",
                operator,
                options,
                widget_id,
                domain_id,
            )
            dt_mgr.load(
                params.granularity,
                params.start,
                params.end,
                params.vars,
            )

            if dt_mgr.state == "UNAVAILABLE":
                raise ERROR_UNAVAILABLE_DATA_TABLE(
                    data_table_id=pub_data_table_vo.data_table_id
                )

            return dt_mgr.response_data(params.sort, params.page)

    @transaction(
        permission="dashboard:PublicDataTable.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "user_projects", "*")
    @convert_model
    def get(
        self, params: PublicDataTableGetRequest
    ) -> Union[PublicDataTableResponse, dict]:
        """Get public data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicDataTableResponse:
        """

        pub_data_table_vo: PublicDataTable = (
            self.pub_data_table_mgr.get_public_data_table(
                params.data_table_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )
        )

        return PublicDataTableResponse(**pub_data_table_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDataTable.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @change_value_by_rule("APPEND", "user_projects", "*")
    @append_query_filter(
        [
            "widget_id",
            "data_table_id",
            "name",
            "state",
            "data_type",
            "source_type",
            "operator",
            "domain_id",
            "workspace_id",
            "project_id",
            "user_projects",
        ]
    )
    @append_keyword_filter(["data_table_id", "name"])
    @convert_model
    def list(
        self, params: PublicDataTableSearchQueryRequest
    ) -> Union[PublicDataTablesResponse, dict]:
        """List public data tables

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'widget_id': 'str',                             # required
                'data_table_id': 'str',
                'name': 'str',
                'data_type': 'str',
                'source_type': 'str',
                'operator': 'str',
                'project_id': 'str',
                'workspace_id': 'str',                          # injected from auth
                'domain_id': 'str',                             # injected from auth (required)
                'user_projects': 'list',                        # injected from auth
            }

        Returns:
            PublicDataTablesResponse:
        """

        query = params.query or {}
        (
            pub_data_table_vos,
            total_count,
        ) = self.pub_data_table_mgr.list_public_data_tables(query)
        pub_data_tables_info = [
            pub_data_table_vo.to_dict() for pub_data_table_vo in pub_data_table_vos
        ]
        return PublicDataTablesResponse(
            results=pub_data_tables_info, total_count=total_count
        )



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PublicDataTable(MongoModel):
    data_table_id = StringField(max_length=40, generate_id="public-dt", unique=True)
    name = StringField(max_length=100, default=None, null=True)
    state = StringField(
        max_length=40, default="AVAILABLE", choices=("AVAILABLE", "UNAVAILABLE")
    )
    error_message = StringField(max_length=255, default=None, null=True)
    data_type = StringField(max_length=40, choices=("ADDED", "TRANSFORMED"))
    source_type = StringField(max_length=40, default=None, null=True)
    operator = StringField(max_length=40, default=None, null=True)
    options = DictField(required=True, default=None)
    tags = DictField(default=None)
    labels_info = DictField(default=None)
    data_info = DictField(default=None)
    dashboard_id = StringField(max_length=40)
    widget_id = StringField(max_length=40)
    resource_group = StringField(
        max_length=40, choices=("DOMAIN", "WORKSPACE", "PROJECT")
    )
    project_id = StringField(max_length=40)
    workspace_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "state",
            "error_message",
            "options",
            "tags",
            "labels_info",
            "data_info",
            "project_id",
            "workspace_id",
        ],
        "minimal_fields": [
            "data_table_id",
            "name",
            "state",
            "data_type",
            "source_type",
            "operator",
            "dashboard_id",
            "widget_id",
            "resource_group",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
        "change_query_keys": {"user_projects": "project_id"},
        "ordering": ["name"],
        "indexes": [
            "name",
            "state",
            "data_type",
            "source_type",
            "operator",
            "dashboard_id",
            "widget_id",
            "resource_group",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
    }



from spaceone.core.error import *


class ERROR_NOT_SUPPORTED_VERSION(ERROR_INVALID_ARGUMENT):
    _message = "Widget does not create in old dashboard version. (dashboard_version = {version})"


class ERROR_NOT_ALLOWED_SHARE(ERROR_INVALID_ARGUMENT):
    _message = (
        "Dashboard the belongs to folder cannot be shared. (folder_id = {folder_id})"
    )


class ERROR_NOT_ALLOWED_UNSHARE(ERROR_INVALID_ARGUMENT):
    _message = (
        "Dashboard the belongs to folder cannot be unshared. (folder_id = {folder_id})"
    )



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.manager.private_folder_manager import PrivateFolderManager
from spaceone.dashboard.manager.identity_manager import IdentityManager
from spaceone.dashboard.model.private_folder.request import *
from spaceone.dashboard.model.private_folder.response import *
from spaceone.dashboard.model.private_folder.database import PrivateFolder

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PrivateFolderService(BaseService):
    resource = "PrivateFolder"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pri_folder_mgr = PrivateFolderManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="dashboard:PrivateFolder.write",
        role_types=["USER"],
    )
    @convert_model
    def create(
        self, params: PrivateFolderCreateRequest
    ) -> Union[PrivateFolderResponse, dict]:
        """Create private folder

        Args:
            params (dict): {
                'name': 'str',                  # required
                'tags': 'dict',
                'dashboards': 'list',
                'workspace_id': 'str',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateFolderResponse:
        """

        if params.workspace_id:
            self.identity_mgr.check_workspace(params.workspace_id, params.domain_id)

        pri_folder_vo = self.pri_folder_mgr.create_private_folder(params.dict())

        # Create child dashboards
        if params.dashboards:
            pass

        return PrivateFolderResponse(**pri_folder_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateFolder.write",
        role_types=["USER"],
    )
    @convert_model
    def update(
        self, params: PrivateFolderUpdateRequest
    ) -> Union[PrivateFolderResponse, dict]:
        """Update private folder

        Args:
            params (dict): {
                'folder_id': 'str',   # required
                'name': 'str',
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateFolderResponse:
        """

        pri_folder_vo = self.pri_folder_mgr.get_private_folder(
            params.folder_id, params.domain_id, params.user_id
        )

        pri_folder_vo = self.pri_folder_mgr.update_private_folder_by_vo(
            params.dict(exclude_unset=True), pri_folder_vo
        )

        return PrivateFolderResponse(**pri_folder_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateFolder.write",
        role_types=["USER"],
    )
    @convert_model
    def delete(self, params: PrivateFolderDeleteRequest) -> None:
        """Delete private folder

        Args:
            params (dict): {
                'folder_id': 'str',   # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_folder_vo = self.pri_folder_mgr.get_private_folder(
            params.folder_id, params.domain_id, params.user_id
        )

        self.pri_folder_mgr.delete_private_folder_by_vo(pri_folder_vo)

    @transaction(
        permission="dashboard:PrivateFolder.read",
        role_types=["USER"],
    )
    @convert_model
    def get(
        self, params: PrivateFolderGetRequest
    ) -> Union[PrivateFolderResponse, dict]:
        """Get private folder

        Args:
            params (dict): {
                'folder_id': 'str',   # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateFolderResponse:
        """

        pri_folder_vo = self.pri_folder_mgr.get_private_folder(
            params.folder_id, params.domain_id, params.user_id
        )

        return PrivateFolderResponse(**pri_folder_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateFolder.read",
        role_types=["USER"],
    )
    @append_query_filter(["folder_id", "name", "domain_id", "workspace_id", "user_id"])
    @append_keyword_filter(["folder_id", "name"])
    @convert_model
    def list(
        self, params: PrivateFolderSearchQueryRequest
    ) -> Union[PrivateFoldersResponse, dict]:
        """List private folders

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'folder_id': 'str',
                'name': 'str',
                'user_id': 'str',                               # injected from auth (required)
                'workspace_id': 'str',
                'domain_id': 'str',                             # injected from auth (required)
            }

        Returns:
            PrivateFoldersResponse:
        """

        query = params.query or {}
        pri_folder_vos, total_count = self.pri_folder_mgr.list_private_folders(query)
        pri_dashboards_info = [
            pri_folder_vo.to_dict() for pri_folder_vo in pri_folder_vos
        ]
        return PrivateFoldersResponse(
            results=pri_dashboards_info, total_count=total_count
        )

    @transaction(
        permission="dashboard:PrivateFolder.read",
        role_types=["USER"],
    )
    @append_query_filter(["domain_id", "user_id"])
    @append_keyword_filter(["folder_id", "name"])
    @convert_model
    def stat(self, params: PrivateFolderStatQueryRequest) -> dict:
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'user_id': 'str',                                       # injected from auth (required)
                'domain_id': 'str'                                      # injected from auth (required)
            }

        Returns:
            dict:

        """

        query = params.query or {}
        return self.pri_folder_mgr.stat_private_folders(query)



from spaceone.core.error import *


class ERROR_NOT_SUPPORTED_SOURCE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = "Data table does not support source type. (source_type = {source_type})"


class ERROR_QUERY_OPTION(ERROR_INVALID_ARGUMENT):
    _message = "Query option is invalid. (key = {key})"


class ERROR_EMPTY_DATA_FIELD(ERROR_INVALID_ARGUMENT):
    _message = "Data field is empty. (fields = {fields})"


class ERROR_QUERY_GROUP_BY_OPTION(ERROR_INVALID_ARGUMENT):
    _message = "Query group by option is invalid. (key = {key}, fields={fields})"


class ERROR_NOT_SUPPORTED_OPERATOR(ERROR_INVALID_ARGUMENT):
    _message = "Data table does not support operator. (operator = {operator})"


class ERROR_UNAVAILABLE_DATA_TABLE(ERROR_INVALID_ARGUMENT):
    _message = "Data table is not available. (data_table_id = {data_table_id})"


class ERROR_DUPLICATED_DATA_FIELDS(ERROR_INVALID_ARGUMENT):
    _message = "Data fields for join are duplicated. (fields = {fields})"


class ERROR_NO_FIELDS_TO_JOIN(ERROR_INVALID_ARGUMENT):
    _message = "There is no fields to join."


class ERROR_NOT_GLOBAL_VARIABLE_KEY(ERROR_INVALID_ARGUMENT):
    _message = "The variable is not a global variable. (global_variable_key = {global_variable_key})"


class ERROR_NO_FIELDS_TO_GLOBAL_VARIABLES(ERROR_INVALID_ARGUMENT):
    _message = "There is no fields in GLOBAL_VARIABLES. (vars = {vars})"


class ERROR_DUPLICATED_FIELD_NAME(ERROR_INVALID_ARGUMENT):
    _message = (
        "The variable to be created is duplicated. (field = {field}, fields = {fields})"
    )


class ERROR_NOT_ALLOWED_DATA_FIELD(ERROR_INVALID_ARGUMENT):
    _message = "The field name is not allowed data field. (name = {name}, data_fields = {data_fields})"



from typing import Union
from pydantic import BaseModel

__all__ = [
    "PublicWidgetCreateRequest",
    "PublicWidgetUpdateRequest",
    "PublicWidgetDeleteRequest",
    "PublicWidgetLoadRequest",
    "PublicWidgetLoadSumRequest",
    "PublicWidgetGetRequest",
    "PublicWidgetSearchQueryRequest",
]


class PublicWidgetCreateRequest(BaseModel):
    dashboard_id: str
    name: Union[str, None] = None
    state: Union[str, None] = None
    description: Union[str, None] = None
    widget_type: Union[str, None] = None
    size: Union[str, None] = None
    options: Union[dict, None] = None
    data_table_id: Union[int, None] = None
    data_tables: Union[list, None] = None
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetUpdateRequest(BaseModel):
    widget_id: str
    name: Union[str, None] = None
    state: Union[str, None] = None
    description: Union[str, None] = None
    widget_type: Union[str, None] = None
    size: Union[str, None] = None
    options: Union[dict, None] = None
    data_table_id: Union[str, None] = None
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetDeleteRequest(BaseModel):
    widget_id: str
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetGetRequest(BaseModel):
    widget_id: str
    workspace_id: Union[str, list, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetLoadRequest(BaseModel):
    widget_id: str
    granularity: str
    start: str
    end: str
    group_by: Union[list, None] = None
    sort: Union[list, None] = None
    page: Union[dict, None] = None
    vars: Union[dict, None] = None
    workspace_id: Union[str, list, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetLoadSumRequest(BaseModel):
    widget_id: str
    granularity: str
    start: str
    end: str
    vars: Union[dict, None] = None
    workspace_id: Union[str, list, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    dashboard_id: str
    widget_id: Union[str, None] = None
    name: Union[str, None] = None
    project_id: Union[str, None] = None
    workspace_id: Union[str, list, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicWidgetStatQueryRequest(BaseModel):
    query: dict
    workspace_id: Union[str, list, None] = None
    domain_id: str
    user_projects: Union[list, None] = None



from spaceone.api.dashboard.v1 import private_widget_pb2, private_widget_pb2_grpc
from spaceone.core.pygrpc import BaseAPI
from spaceone.dashboard.service.private_widget_service import PrivateWidgetService


class PrivateWidget(BaseAPI, private_widget_pb2_grpc.PrivateWidgetServicer):
    pb2 = private_widget_pb2
    pb2_grpc = private_widget_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        response: dict = pri_widget_svc.create(params)
        return self.dict_to_message(response)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        response: dict = pri_widget_svc.update(params)
        return self.dict_to_message(response)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        pri_widget_svc.delete(params)
        return self.empty()

    def load(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        response: dict = pri_widget_svc.load(params)
        return self.dict_to_message(response)

    def load_sum(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        response: dict = pri_widget_svc.load_sum(params)
        return self.dict_to_message(response)

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        response: dict = pri_widget_svc.get(params)
        return self.dict_to_message(response)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_widget_svc = PrivateWidgetService(metadata)
        response: dict = pri_widget_svc.list(params)
        return self.dict_to_message(response)
